#! python3  # noqa: E265

"""Custom exceptions."""

from collections.abc import Iterable

# standard library
from pathlib import Path

# package
from qgis_deployment_toolbelt.utils.check_image_size import get_image_size


class JobOptionBadName(KeyError):
    """When a job reveives an option which is not part of accepted ones."""

    def __init__(
        self, job_id: str, bad_option_name: str, expected_options_names: Iterable[str]
    ):
        """Initialization method.

        Args:
            job_id (str): job ID
            bad_option_name (str): name of bad option passed
            expected_options_names (Iterable[str]): _description_
        """
        self.message = (
            f"Job: {job_id}. Option '{bad_option_name}' is not valid. "
            f"Valid options are: {','.join(expected_options_names)}"
        )

        super().__init__(self.message)


class JobOptionBadValue(ValueError):
    """When a job's option reveives a value which does not complies with condition."""

    def __init__(
        self,
        job_id: str,
        bad_option_name: str,
        bad_option_value: str,
        condition: str,
        accepted_values: Iterable[str],
    ):
        """Initialization method.

        Args:
            job_id (str): job ID
            bad_option_name (str): name of bad option passed
            bad_option_type (str): name of bad option passed
            condition (str): condition
            accepted_values (Iterable[str]): accepted types of values
        """
        self.message = (
            f"Job: {job_id}. Option '{bad_option_name}' 's value '{bad_option_value}' "
            f"does not comply with condition {condition}. "
            f"Accepted pattern: {','.join(accepted_values)}"
        )

        super().__init__(self.message)


class JobOptionBadValueType(TypeError):
    """When a job reveives an option which is not of the expected type."""

    def __init__(
        self,
        job_id: str,
        bad_option_name: str,
        bad_option_value: str,
        expected_option_type: Iterable[str],
    ):
        """Initialization method.

        Args:
            job_id (str): job ID
            bad_option_name (str): name of bad option passed
            bad_option_value (str): option's value
            expected_option_type (Iterable[str]): accepted types of values
        """
        self.message = (
            f"Job: {job_id}. Option '{bad_option_name}' 's value '{bad_option_value}' "
            f"has an invalid type: {type(bad_option_value)}. "
            f"Valid type/s is/are: {expected_option_type}"
        )

        super().__init__(self.message)


class SplashScreenBadDimensions(Exception):
    """When splash screen image does not comply with recomended dimensions."""

    def __init__(self, image_filepath: Path, profile_name: str | None = None):
        """Initialization method.

        Args:
            image_filepath (Path): path to the selected image to be set as splash screen
            profile_name (str, optional): name of profile. Defaults to None.
        """
        self.message = (
            f"Profile {profile_name} -"
            f"{image_filepath.resolve()} dimensions "
            f"{get_image_size(image_filepath=image_filepath)} do not comply with "
            " dimensions recomended by QGIS for splash screen: 600x300."
        )
        super().__init__(self.message)


class QgisInstallNotFound(Exception):
    """When no QGIS installation are found."""

    def __init__(self):
        """Initialization method."""
        self.message = "No QGIS installation found."
        super().__init__(self.message)



#! python3  # noqa E265

"""
    Usage from the repo root folder:

    .. code-block:: bash
        # for whole tests
        python -m unittest tests.test_qgis_ini_helper
        # for specific test
        python -m unittest tests.test_qgis_ini_helper.TestQgisIniHelper.test_profile_load_from_json_basic
"""

# standard
import tempfile
import unittest
from configparser import ConfigParser
from pathlib import Path

# project
from qgis_deployment_toolbelt.profiles.qgis_ini_handler import QgisIniHelper

# ############################################################################
# ########## Classes #############
# ################################


class TestQgisIniHelper(unittest.TestCase):
    """Test module handling QGIS configuration files (*.ini)."""

    # -- Standard methods --------------------------------------------------------
    @classmethod
    def setUpClass(cls):
        """Executed when module is loaded before any test."""
        cls.configuration_files = sorted(Path("tests/fixtures/").glob("**/QGIS3*.ini"))
        cls.customization_files = sorted(Path("tests/fixtures/").glob("**/QGIS3*.ini"))

    def test_load_profile_config_default(self):
        """Test profile QGIS/QGIS3.ini loader."""
        new_config_file = Path(
            "tests/fixtures/qgis_ini/default_no_customization/QGIS3.ini"
        )

        with tempfile.TemporaryDirectory(
            prefix="qdt_test_ini_file_", ignore_cleanup_errors=True
        ) as tmpdirname:
            tmp_copy = Path(tmpdirname).joinpath(new_config_file.name)
            tmp_copy.write_text(new_config_file.read_text())

            # open temp copy
            ini_config = QgisIniHelper(ini_filepath=tmp_copy)

        self.assertEqual(ini_config.ini_type, "profile_qgis3")
        self.assertFalse(ini_config.is_ui_customization_enabled())

    def test_load_profile_customization_splash_screen(self):
        """Test profile QGIS/QGIS3.ini loader."""
        fixture_ini_file = Path(
            "tests/fixtures/qgis_ini/default_customization/QGISCUSTOMIZATION3.ini"
        )

        with tempfile.TemporaryDirectory(
            prefix="qdt_test_ini_file_", ignore_cleanup_errors=True
        ) as tmpdirname:
            tmp_copy = Path(tmpdirname).joinpath(fixture_ini_file.name)
            tmp_copy.write_text(fixture_ini_file.read_text())

            ini_customization = QgisIniHelper(ini_filepath=tmp_copy)

        self.assertEqual(ini_customization.ini_type, "profile_qgis3customization")
        ini_config = QgisIniHelper(ini_filepath=ini_customization.profile_config_path)
        self.assertEqual(ini_config.ini_type, "profile_qgis3")

    def test_enable_customization(self):
        """Test profile QGIS/QGIS3.ini loader."""
        fixture_ini_file = Path(
            "tests/fixtures/qgis_ini/default_no_customization/QGIS3.ini"
        )

        with tempfile.TemporaryDirectory(
            prefix="qdt_test_ini_file_", ignore_cleanup_errors=True
        ) as tmpdirname:
            tmp_copy = Path(tmpdirname).joinpath(fixture_ini_file.name)
            tmp_copy.write_text(fixture_ini_file.read_text())

            ini_config = QgisIniHelper(ini_filepath=tmp_copy)
            self.assertFalse(ini_config.is_ui_customization_enabled())

            ini_config.set_ui_customization_enabled(switch=True)
            self.assertTrue(ini_config.is_ui_customization_enabled())

            ini_config.set_ui_customization_enabled(switch=False)
            self.assertFalse(ini_config.is_ui_customization_enabled())

    def test_enable_customization_on_unexisting_file(self):
        """Test profile QGIS/QGIS3.ini loader."""

        with tempfile.TemporaryDirectory(
            prefix="qdt_test_ini_file_", ignore_cleanup_errors=True
        ) as tmpdirname:
            unexisting_config_file = Path(tmpdirname).joinpath("Imaginary_QGIS3.ini")
            self.assertFalse(unexisting_config_file.exists())

            ini_config = QgisIniHelper(
                ini_filepath=unexisting_config_file, ini_type="profile_qgis3"
            )
            self.assertFalse(ini_config.is_ui_customization_enabled())
            self.assertFalse(unexisting_config_file.exists())

            ini_config.set_ui_customization_enabled(switch=True)
            self.assertTrue(ini_config.is_ui_customization_enabled())
            self.assertTrue(unexisting_config_file.exists())

            ini_config.set_ui_customization_enabled(switch=False)
            self.assertFalse(ini_config.is_ui_customization_enabled())
            self.assertTrue(unexisting_config_file.exists())

    def test_splash_already_set(self):
        fake_config = (
            "[Customization]\nsplashpath=/home/$USER/.share/QGIS/QGIS3/images/"
        )
        with tempfile.TemporaryDirectory(
            prefix="qdt_test_ini_file_", ignore_cleanup_errors=True
        ) as tmpdirname:
            tmp_ini_customization = Path(tmpdirname).joinpath(
                "customization_with_splashpath.ini"
            )
            tmp_ini_customization.parent.mkdir(parents=True, exist_ok=True)
            tmp_ini_customization.write_text(fake_config)

            qini_helper = QgisIniHelper(
                ini_filepath=tmp_ini_customization,
                ini_type="profile_qgis3customization",
            )
            self.assertTrue(qini_helper.is_splash_screen_set())

    def test_splash_already_set_but_empty(self):
        fake_config = "[Customization]\nsplashpath="

        with tempfile.TemporaryDirectory(
            prefix="qdt_test_ini_file_", ignore_cleanup_errors=True
        ) as tmpdirname:
            tmp_ini_customization = Path(tmpdirname).joinpath(
                "customization_with_splashpath_empty.ini"
            )
            tmp_ini_customization.parent.mkdir(parents=True, exist_ok=True)
            tmp_ini_customization.write_text(fake_config)

            qini_helper = QgisIniHelper(
                ini_filepath=tmp_ini_customization,
                ini_type="profile_qgis3customization",
            )
            self.assertFalse(qini_helper.is_splash_screen_set())

    def test_splash_not_set(self):
        fake_config = "[Customization]\nMenu\\mDatabaseMenu=false"
        with tempfile.TemporaryDirectory(
            prefix="qdt_test_ini_file_", ignore_cleanup_errors=True
        ) as tmpdirname:
            tmp_ini_customization = Path(tmpdirname).joinpath(
                "customization_with_no_splashpath_set.ini"
            )
            tmp_ini_customization.parent.mkdir(parents=True, exist_ok=True)
            tmp_ini_customization.write_text(fake_config)

            qini_helper = QgisIniHelper(
                ini_filepath=tmp_ini_customization,
                ini_type="profile_qgis3customization",
            )
            self.assertFalse(qini_helper.is_splash_screen_set())

    def test_disable_splash_already_not_set(self):
        fake_config = "[Customization]\nMenu\\mDatabaseMenu=false"
        with tempfile.TemporaryDirectory(
            prefix="qdt_test_ini_file_", ignore_cleanup_errors=True
        ) as tmpdirname:
            tmp_ini_customization = Path(tmpdirname).joinpath(
                "set_splash_path_on_customization_with_no_splashpath_set.ini"
            )
            tmp_ini_customization.parent.mkdir(parents=True, exist_ok=True)
            tmp_ini_customization.write_text(fake_config)

            qini_helper = QgisIniHelper(
                ini_filepath=tmp_ini_customization,
                ini_type="profile_qgis3customization",
            )
            self.assertFalse(qini_helper.set_splash_screen(switch=False))

    def test_disable_splash_on_unexisting_file(self):
        with tempfile.TemporaryDirectory(
            prefix="qdt_test_ini_file_", ignore_cleanup_errors=True
        ) as tmpdirname:
            not_existing_ini_customization = Path(tmpdirname).joinpath(
                "not_existing_customization.ini"
            )

            qini_helper = QgisIniHelper(
                ini_filepath=not_existing_ini_customization,
                ini_type="profile_qgis3customization",
            )
            self.assertFalse(qini_helper.set_splash_screen(switch=False))

            not_existing_ini_config = Path("QGIS3.ini")

            qini_helper = QgisIniHelper(
                ini_filepath=not_existing_ini_config, ini_type="profile_qgis3"
            )
            self.assertFalse(qini_helper.set_splash_screen(switch=False))

    def test_merge_existing_file(self):
        """Test merge of INI file with backup value"""
        dst_config = "[Section]\nvalue=initial_value"
        src_config = "[Section]\nvalue=new_value"
        with tempfile.TemporaryDirectory(
            prefix="qdt_test_ini_file_", ignore_cleanup_errors=True
        ) as tmpdirname:
            dst_ini_path = Path(tmpdirname).joinpath("dest.ini")
            dst_ini_path.parent.mkdir(parents=True, exist_ok=True)
            dst_ini_path.write_text(dst_config)

            dst_ini = QgisIniHelper(
                ini_filepath=dst_ini_path,
                ini_type="profile_qgis3customization",
            )

            src_ini_path = Path(tmpdirname).joinpath("src.ini")
            src_ini_path.parent.mkdir(parents=True, exist_ok=True)
            src_ini_path.write_text(src_config)

            src_ini = QgisIniHelper(
                ini_filepath=src_ini_path,
                ini_type="profile_qgis3customization",
            )

            src_ini.merge_to(dst_ini)

            cfg_parser = ConfigParser()
            cfg_parser.read(dst_ini_path)
            self.assertIn("QDT_backup_Section", cfg_parser)
            self.assertEqual("initial_value", cfg_parser["QDT_backup_Section"]["value"])
            self.assertIn("Section", cfg_parser)
            self.assertEqual("new_value", cfg_parser["Section"]["value"])

    def test_merge_not_existing_file(self):
        """Test merge of INI file with not existing file"""
        src_config = "[Section]\nvalue=new_value"
        with tempfile.TemporaryDirectory(
            prefix="qdt_test_ini_file_", ignore_cleanup_errors=True
        ) as tmpdirname:
            src_ini_path = Path(tmpdirname).joinpath("src.ini")
            src_ini_path.parent.mkdir(parents=True, exist_ok=True)
            src_ini_path.write_text(src_config)

            src_ini = QgisIniHelper(
                ini_filepath=src_ini_path,
                ini_type="profile_qgis3customization",
            )

            empty_ini_path = Path(tmpdirname).joinpath("empty.ini")
            empty_ini = QgisIniHelper(
                ini_filepath=empty_ini_path,
                ini_type="profile_qgis3customization",
            )
            src_ini.merge_to(empty_ini)

            cfg_parser = ConfigParser()
            cfg_parser.read(empty_ini_path)
            self.assertNotIn("QDT_backup_Section", cfg_parser)
            self.assertIn("Section", cfg_parser)
            self.assertEqual("new_value", cfg_parser["Section"]["value"])


# ############################################################################
# ####### Stand-alone run ########
# ################################
if __name__ == "__main__":
    unittest.main()



#! python3  # noqa: E265
# Copyright 2023 Julien Moura (Oslandia)
# SPDX-License-Identifier: Apache-2.0

"""
    Main command-line.
"""

# #############################################################################
# ########## Libraries #############
# ##################################

# Standard library
import argparse
import logging
import sys
from os import environ

# submodules
from qgis_deployment_toolbelt.__about__ import (
    __author__,
    __summary__,
    __title__,
    __title_clean__,
    __uri_homepage__,
    __version__,
)
from qgis_deployment_toolbelt.commands.cmd_rules_context import (
    parser_rules_context_export,
)
from qgis_deployment_toolbelt.commands.deployment import parser_main_deployment
from qgis_deployment_toolbelt.commands.upgrade import parser_upgrade
from qgis_deployment_toolbelt.utils.journalizer import configure_logger

# ############################################################################
# ########## FUNCTIONS ###########
# ################################


def add_common_arguments(
    parser_to_update: argparse.ArgumentParser,
    add_verbosity: bool = True,
    add_proxy: bool = True,
):
    """Apply common arguments to an existing parser.

    Args:
        parser_to_update (argparse.ArgumentParser): parser to which arguments need to be added
        add_verbosity (bool, optional): if enabled, add --verbose. Defaults to True.
        add_proxy (bool, optional): if enabled, adds --proxy-http. Defaults to True.

    Returns:
        argparse.ArgumentParser: parser with added options
    """
    if add_verbosity:
        parser_to_update.add_argument(
            "-v",
            "--verbose",
            action="count",
            default=1,
            dest="verbosity",
            help="Verbosity level. None = WARNING, -v = INFO, -vv = DEBUG. Can be set with "
            "QDT_LOGS_LEVEL environment variable and logs location with QDT_LOGS_DIR.",
        )

    if add_proxy:
        parser_to_update.add_argument(
            "--proxy-http",
            default=None,
            dest="proxy_http",
            help="Option to specify an HTTP proxy in the form: "
            "scheme://[user:passwd@]proxy.server:port",
            metavar="QDT_PROXY_HTTP",
            type=str,
        )

    return parser_to_update


def set_default_subparser(
    parser_to_update: argparse.ArgumentParser,
    default_subparser_name: str,
    args: list | None = None,
):
    """Set a default subparser to a parent parser. Call after setup and just before
        parse_args().
        See: <https://stackoverflow.com/questions/5176691/argparse-how-to-specify-a-default-subcommand>

    Args:
        parser_to_update (argparse.ArgumentParser): parent parser to add
        default_subparser_name (str): name of the subparser to call by default
        args (list, optional): if set is the argument list handed to parse_args().
        Defaults to None.
    """
    subparser_found = False
    for arg in sys.argv[1:]:
        if arg in [
            "-h",
            "--help",
            "--version",
            "--no-logfile",
        ]:  # ignore main parser args
            break

    else:
        for x in parser_to_update._subparsers._actions:
            if not isinstance(x, argparse._SubParsersAction):
                continue
            for sp_name in x._name_parser_map.keys():
                if sp_name in sys.argv[1:]:
                    subparser_found = True
        if not subparser_found:
            # insert default in first position, this implies no
            # global options without a sub_parsers specified
            if args is None:
                sys.argv.insert(1, default_subparser_name)
            else:
                args.insert(0, default_subparser_name)


# ############################################################################
# ########## MAIN ################
# ################################
def main(in_args: list[str] | None = None):
    """Main CLI entrypoint.

    Args:
        in_args (List[str], optional): list of command-line arguments. Defaults to None.
    """
    # create the top-level parser
    main_parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog=f"Developed by: {__author__}\nDocumentation: {__uri_homepage__}",
        description=f"{__title__} {__version__} - {__summary__}",
        argument_default=argparse.SUPPRESS,
    )

    # -- ROOT ARGUMENTS --
    main_parser.add_argument(
        "--no-logfile",
        default=True,
        action="store_false",
        dest="opt_logfile_disabled",
        help="Disable log file. Log files are usually created, rotated and stored in the"
        "folder set by QDT_LOGS_DIR.",
    )

    main_parser.add_argument(
        "--version",
        action="version",
        version=__version__,
        help="Display CLI version",
    )

    add_common_arguments(main_parser, add_verbosity=True, add_proxy=True)
    # -- SUB-COMMANDS --
    subparsers = main_parser.add_subparsers(title="Sub-commands", dest="command")

    # Main logic
    subcmd_deployment = subparsers.add_parser(
        "deploy",
        help="QDT's main logic: run the deployment's scenario.",
        formatter_class=main_parser.formatter_class,
        prog="deployment",
    )
    add_common_arguments(subcmd_deployment)
    parser_main_deployment(subcmd_deployment)

    # Rules context
    subcmd_rules_context = subparsers.add_parser(
        "export-rules-context",
        help="Export QDT rules context taking into account the local environment to "
        "help rules writing.",
        formatter_class=main_parser.formatter_class,
        prog="rules-context-export",
    )
    add_common_arguments(subcmd_rules_context)
    parser_rules_context_export(subcmd_rules_context)

    # Upgrader
    subcmd_upgrade = subparsers.add_parser(
        "upgrade",
        aliases=["auto-update", "update"],
        help="Check if a new version of QDT is available and download it locally.",
        formatter_class=main_parser.formatter_class,
        prog="upgrade",
    )
    add_common_arguments(subcmd_upgrade)
    parser_upgrade(subcmd_upgrade)

    # -- PARSE ARGS --
    set_default_subparser(parser_to_update=main_parser, default_subparser_name="deploy")

    # just get passed args
    args = main_parser.parse_args(in_args)

    # proxy configuration
    if args.proxy_http:
        environ["QDT_PROXY_HTTP"] = args.proxy_http

    # log configuration
    if args.opt_logfile_disabled:
        configure_logger(
            verbosity=args.verbosity, logfile=f"{__title_clean__}_{__version__}.log"
        )
    else:
        configure_logger(verbosity=args.verbosity)

    # add the handler to the root logger
    logger = logging.getLogger(__title_clean__)
    logger.debug(f"Log level set: {logging.getLevelName(args.verbosity)}")

    # -- RUN LOGIC --
    if hasattr(args, "func"):
        args.func(args)


# #############################################################################
# ##### Stand alone program ########
# ##################################

if __name__ == "__main__":
    """Standalone execution."""
    # launch cli
    main()  # required by unittest



#! python3  # noqa E265

"""
    Usage from the repo root folder:

    .. code-block:: bash
        # for whole tests
        python -m unittest tests.test_utils_url_helpers
        # for specific test
        python -m unittest tests.test_utils_url_helpers.TestUtilsUrlHelpers.test_check_str_is_url
"""


# standard library
import unittest
from pathlib import Path

# project
from qgis_deployment_toolbelt.__about__ import __uri__
from qgis_deployment_toolbelt.utils.url_helpers import check_str_is_url

# ############################################################################
# ########## Classes #############
# ################################


class TestUtilsUrlHelpers(unittest.TestCase):
    """Test URL helpers."""

    def test_check_str_is_url(self):
        """Test function that determines if a str or Path is a valid URL."""
        self.assertTrue(check_str_is_url(input_str=__uri__))
        self.assertTrue(
            check_str_is_url(input_str="ftp://fakeftp:21", ref_shemes=("ftp", "http")),
        )
        self.assertFalse(check_str_is_url(input_str=Path(__uri__), raise_error=False))
        self.assertFalse(check_str_is_url(input_str=Path(__file__), raise_error=False))

        with self.assertRaises((TypeError, ValueError)):
            check_str_is_url(input_str=Path(__file__), raise_error=True)


# ############################################################################
# ####### Stand-alone run ########
# ################################
if __name__ == "__main__":
    unittest.main()



#! python3  # noqa: E265

"""
    Test CLI's main command.

    Author: Julien Moura (Oslandia)
"""

# #############################################################################
# ########## Libraries #############
# ##################################


# Standard library
import tempfile
from os import environ, getenv
from pathlib import Path

# 3rd party
import pytest

# project
from qgis_deployment_toolbelt import __about__, cli
from qgis_deployment_toolbelt.profiles.qdt_profile import QdtProfile

# #############################################################################
# ######## Globals #################
# ##################################

sample_scenario_good: Path = Path(
    "tests/fixtures/scenarios/good_scenario_sample.qdt.yml"
)
sample_scenario_good_with_unexisting_jobs: Path = Path(
    "tests/fixtures/scenarios/good_scenario_with_unexisting_jobs.qdt.yml"
)
sample_scenario_good_splash_removal: Path = Path(
    "tests/fixtures/scenarios/good_scenario_splash_screen_remove.qdt.yml"
)

sample_scenario_false: Path = Path(
    "tests/fixtures/scenarios/false_scenario_sample.qdt.yml"
)
sample_scenario_imaginary: Path = Path(
    "tests/fixtures/scenarios/imaginary_scenario_sample.qdt.yml"
)

good_scenarios = [
    ["deploy", f"--scenario={scenario_path.resolve()}"]
    for scenario_path in Path("tests/fixtures/scenarios/").glob("good_*.qdt.yml")
]

# #############################################################################
# ######## Classes #################
# ##################################


@pytest.mark.parametrize("option", ("-h", "--help"))
def test_cli_help(capsys, option):
    """Test CLI help."""
    with pytest.raises(SystemExit):
        cli.main([option])

    out, err = capsys.readouterr()

    assert (
        f"{__about__.__title__} {__about__.__version__} - {__about__.__summary__}"
        in out
    )
    assert err == ""


@pytest.mark.parametrize("option", (["--version"]))
def test_cli_version(capsys, option):
    """Test CLI version."""
    with pytest.raises(SystemExit):
        cli.main([option])

    out, err = capsys.readouterr()

    assert f"{__about__.__version__}\n" == out

    assert err == ""


@pytest.mark.parametrize("option", good_scenarios)
def test_main_run(capsys, option):
    """Test main cli command"""

    with tempfile.TemporaryDirectory(
        prefix="QDT_test_cli_run_",
        ignore_cleanup_errors=True,
    ) as tmpdirname:
        # customize QDT working folder and profiles destination folder
        tmp_dir = Path(tmpdirname).joinpath(f"scenario_{good_scenarios.index(option)}")
        environ["QDT_LOCAL_WORK_DIR"] = (
            f"{Path(tmpdirname).joinpath('qdt_working_folder').resolve()}"
        )
        environ["QGIS_CUSTOM_CONFIG_PATH"] = f"{tmp_dir.resolve()}"

        assert getenv("QGIS_CUSTOM_CONFIG_PATH") is not None

        with pytest.raises(SystemExit):
            cli.main(option)

        _, err = capsys.readouterr()
        assert err == ""

        # checks
        created_profiles = [
            QdtProfile.from_json(profile_json_path=f, profile_folder=f.parent)
            for f in tmp_dir.glob("**/profile.json")
        ]

        assert Path(getenv("QGIS_CUSTOM_CONFIG_PATH")).is_dir()
        assert len(created_profiles) > 0

    # clean up environment vars
    environ.pop("QGIS_CUSTOM_CONFIG_PATH")


def test_main_run_as_admin(capsys):
    """Test main cli command on scenario requiring admin rights"""
    with pytest.raises(SystemExit):
        cli.main(
            [
                "deploy",
                f"--scenario={Path('tests/fixtures/scenarios/scenario_sample_as_admin.qdt.yml').resolve()}",
            ]
        )

    _, err = capsys.readouterr()
    assert err == ""


def test_main_run_unexising_jobs(capsys):
    """Test main cli command"""
    with pytest.raises(SystemExit):
        cli.main(
            [
                "deploy",
                f"--scenario={str(sample_scenario_good_with_unexisting_jobs.resolve())}",
            ]
        )

    _, err = capsys.readouterr()
    assert err == ""


def test_main_run_failed(capsys):
    """Test main cli command"""
    with pytest.raises(FileExistsError):
        cli.main(["deploy", f"--scenario={str(sample_scenario_false.resolve())}"])

    _, err = capsys.readouterr()
    assert err == ""


def test_main_run_removing_splash(capsys):
    """Test main cli command"""
    with tempfile.TemporaryDirectory(
        prefix="qdt_test_cli_main_", ignore_cleanup_errors=True
    ) as tmpdirname:
        environ["QGIS_CUSTOM_CONFIG_PATH"] = tmpdirname
        with pytest.raises(SystemExit):
            cli.main(
                [
                    "deploy",
                    f"--scenario={sample_scenario_good_splash_removal.resolve()}",
                ]
            )

        _, err = capsys.readouterr()
        assert err == ""



#! python3  # noqa: E265

"""
    Synchronize plugins between downloaded and installed profiles.

    Author: Julien Moura (https://github.com/guts)
"""


# #############################################################################
# ########## Libraries #############
# ##################################

# Standard library
import logging
from pathlib import Path
from shutil import unpack_archive

# package
from qgis_deployment_toolbelt.jobs.generic_job import GenericJob
from qgis_deployment_toolbelt.plugins.plugin import QgisPlugin
from qgis_deployment_toolbelt.profiles.qdt_profile import QdtProfile

# #############################################################################
# ########## Globals ###############
# ##################################

# logs
logger = logging.getLogger(__name__)


# #############################################################################
# ########## Classes ###############
# ##################################


class JobPluginsSynchronizer(GenericJob):
    """
    Job to download and synchronize plugins.
    """

    ID: str = "qplugins-synchronizer"
    OPTIONS_SCHEMA: dict = {
        "action": {
            "type": str,
            "required": False,
            "default": "create_or_restore",
            "possible_values": ("create", "create_or_restore", "remove"),
            "condition": "in",
        },
        "profile_ref": {
            "type": str,
            "required": False,
            "default": "installed",
            "possible_values": ("downloaded", "installed"),
            "condition": None,
        },
        "source": {
            "type": str,
            "required": False,
            "default": None,
            "possible_values": None,
            "condition": None,
        },
    }

    def __init__(self, options: dict) -> None:
        """Instantiate the class.

        :param dict options:  job options.
        """
        super().__init__()
        self.options: dict = self.validate_options(options)

        # where QDT downloads plugins
        self.qdt_plugins_folder.mkdir(exist_ok=True, parents=True)
        logger.info(f"QDT plugins folder: {self.qdt_plugins_folder}")

        # which profile.json file to use
        if self.options.get("profile_ref") == "installed":
            self.profiles_path = self.qgis_profiles_path
            logger.debug(
                "Using plugins listed into profile.json files found into profiles "
                "already installed under the QGIS3 user folder: "
                f"{self.profiles_path.resolve()}"
            )
        else:
            self.profiles_path = self.qdt_working_folder
            logger.debug(
                "Using plugins listed into profile.json files found into profiles "
                f"downloaded under the QDT local folder: {self.profiles_path.resolve()}"
            )

    def run(self) -> None:
        """Execute job logic."""
        # local vars
        profile_plugins_to_create: list[tuple[QdtProfile, QgisPlugin, Path]] = []
        profile_plugins_to_restore = []
        profile_plugins_to_upgrade = []

        # get profiles, downloaded or installed
        qdt_profiles = self.filter_profiles_folder(
            start_parent_folder=self.profiles_path
        )

        if qdt_profiles is None:
            logger.error(
                f"No QGIS profile found in {self.options.get('profile_ref')} folder: "
                f"{self.profiles_path}"
            )
            return

        for qdt_profile in qdt_profiles:
            # determine folder
            if self.options.get("installed"):
                profile_plugins_folder = qdt_profile.folder / "python/plugins"
            else:
                profile_plugins_folder = qdt_profile.path_in_qgis / "python/plugins"

            # parse plugins in profile
            for expected_plugin in qdt_profile.plugins:
                # expected_plugin = expected version to be installed into the profile

                # is the plugin downloaded
                plugin_downloaded_zip_source = (
                    self.qdt_plugins_folder / f"{expected_plugin.id_with_version}.zip"
                )
                if not plugin_downloaded_zip_source.is_file():
                    logger.warning(
                        f"Profile {qdt_profile.name} - "
                        f"Plugin {expected_plugin.name} version "
                        f"{expected_plugin.version} should be installed but its "
                        f"archive is not found: {plugin_downloaded_zip_source}"
                    )
                    continue

                # check if the plugin is already installed or not
                plugin_installed_folder = Path(
                    profile_plugins_folder, expected_plugin.installation_folder_name
                )
                if not plugin_installed_folder.is_dir():
                    logger.debug(
                        f"Profile {qdt_profile.name} - "
                        f"Plugin {expected_plugin.name} is not present. It will be added."
                    )
                    profile_plugins_to_create.append(
                        (
                            qdt_profile,
                            expected_plugin,
                            plugin_downloaded_zip_source,
                        )
                    )
                    continue

                # if the plugin is already present into the profile
                plugin_installed: QgisPlugin = QgisPlugin.from_plugin_folder(
                    input_plugin_folder=plugin_installed_folder
                )

                # if the installed plugin has the same version, don't touch anything
                if plugin_installed.version == expected_plugin.version:
                    logger.debug(
                        f"Profile {qdt_profile.name} - "
                        f"Plugin {expected_plugin.name} is already installed "
                        f"with the expected version: {expected_plugin.version}"
                    )
                    continue

                # if verisons are different
                if plugin_installed.is_older_than(expected_plugin):
                    logger.info(
                        f"Profile {qdt_profile.name} - "
                        f"Plugin {expected_plugin.name} is already installed "
                        f"but in an older version: {plugin_installed.version} < "
                        f"{expected_plugin.version}. It will be upgraded."
                    )
                    profile_plugins_to_upgrade.append(
                        (qdt_profile, expected_plugin, plugin_downloaded_zip_source)
                    )

        # log parse results
        if not any(
            (
                profile_plugins_to_create,
                profile_plugins_to_restore,
                profile_plugins_to_upgrade,
            )
        ):
            logger.info(
                "Every plugins are up to date in the "
                f"{len(qdt_profiles)} profiles parsed."
            )
        else:
            self.install_plugin_into_profile(profile_plugins_to_create)
            self.install_plugin_into_profile(profile_plugins_to_upgrade)

        logger.debug(f"Job {self.ID} ran successfully.")

    # -- INTERNAL LOGIC ------------------------------------------------------
    def install_plugin_into_profile(
        self, list_plugins_to_profiles: list[tuple[QdtProfile, QgisPlugin, Path]]
    ):
        """Unzip downloaded plugins into the matching profiles.

        Args:
            list_plugins_to_profiles (List[Tuple[QdtProfile, QgisPlugin, Path]]): list \
                of tuples containing the target profile, the plugin object and the ZIP path.
        """
        for profile, plugin, source_path in list_plugins_to_profiles:
            if self.options.get("installed"):
                profile_plugins_folder = profile.folder / "python/plugins"
            else:
                profile_plugins_folder = profile.path_in_qgis / "python/plugins"

            # make sure destination folder exists
            profile_plugins_folder.mkdir(parents=True, exist_ok=True)

            unpack_archive(filename=source_path, extract_dir=profile_plugins_folder)

            logger.info(
                f"Profile {profile.name} - "
                f"Plugin {plugin.name} {plugin.version} has been unzipped from "
                f"{source_path} to {profile_plugins_folder}"
            )


# #############################################################################
# ##### Stand alone program ########
# ##################################

if __name__ == "__main__":
    """Standalone execution."""
    pass



#! python3  # noqa: E265


# ############################################################################
# ########## IMPORTS #############
# ################################

# standard library
import logging
import ssl
import warnings
from os import getenv
from pathlib import Path

# 3rd party
import truststore
from requests import Response, Session
from requests.adapters import HTTPAdapter
from requests.exceptions import ConnectionError, HTTPError
from requests.utils import requote_uri
from urllib3.exceptions import InsecureRequestWarning

# package
from qgis_deployment_toolbelt.__about__ import __title_clean__, __version__
from qgis_deployment_toolbelt.utils.formatters import convert_octets
from qgis_deployment_toolbelt.utils.proxies import get_proxy_settings
from qgis_deployment_toolbelt.utils.str2bool import str2bool

# ############################################################################
# ########## GLOBALS #############
# ################################

# logs
logger = logging.getLogger(__name__)

if not str2bool(getenv("QDT_SSL_VERIFY", True)):
    warnings.filterwarnings("ignore", category=InsecureRequestWarning)
    logger.warning(
        "SSL warnings (InsecureRequestWarning) are explicitly disabled through "
        "environment variable 'QDT_SSL_VERIFY'. HTTPS request will be unverified. "
        "Adding certificate verification is strongly advised. "
        "See: https://urllib3.readthedocs.io/en/latest/advanced-usage.html#tls-warnings"
    )


# ############################################################################
# ########## CLASSES #############
# ################################


class TruststoreAdapter(HTTPAdapter):
    """Custom HTTP transport adapter made to use local trust store.

    Source: <https://stackoverflow.com/a/78265028/2556577>
    Documentation: <https://requests.readthedocs.io/en/latest/user/advanced/#transport-adapters>
    """

    def init_poolmanager(
        self, connections: int, maxsize: int, block: bool = False
    ) -> None:
        """Initializes a urllib3 PoolManager.

        Args:
            connections (int): number of urllib3 connection pools to cache.
            maxsize (int): maximum number of connections to save in the pool.
            block (bool, optional): Block when no free connections are available.. Defaults to False.

        """
        ctx = truststore.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
        return super().init_poolmanager(connections, maxsize, block, ssl_context=ctx)


# ############################################################################
# ########## FUNCTIONS ###########
# ################################


def download_remote_file_to_local(
    remote_url_to_download: str,
    local_file_path: Path,
    user_agent: str = f"{__title_clean__}/{__version__}",
    content_type: str | None = None,
    chunk_size: int = 8192,
    timeout: tuple[int, int] = (800, 800),
    use_stream: bool = True,
) -> Path:
    """Check if the local index file exists. If not, download the search index from \
        remote URL. If it does exist, check if it has been modified.

    Args:
        remote_url_to_download (str): remote URL of the search index
        local_file_path (Path): local path to the index file
        user_agent (str, optional): user agent to use to perform the request. Defaults \
            to f"{__title_clean__}/{__version__}".
        content_type (str | None, optional): HTTP content-type. Defaults to None.
        chunk_size (int, optional): size of each chunk to read and write in bytes. \
            Defaults to 8192.
        timeout (tuple[int, int], optional): custom timeout (request, response). \
            Defaults to (800, 800).
        use_stream (bool, optional): Option to enable/disable streaming download. \
            Defaults to True.

    Returns:
        Path: path to the local file (should be the same as local_file_path)
    """
    # check if file exists
    if local_file_path.exists():
        logger.info(f"{local_file_path} already exists. It's about to be replaced.")
        local_file_path.unlink(missing_ok=True)

    # make sure parents folder exist
    local_file_path.parent.mkdir(parents=True, exist_ok=True)

    # headers
    headers = {"User-Agent": user_agent}
    if content_type:
        headers["Accept"] = content_type

    try:
        with Session() as dl_session:
            dl_session.headers.update(headers)
            dl_session.proxies.update(get_proxy_settings())
            dl_session.verify = str2bool(getenv("QDT_SSL_VERIFY", True))

            # handle local system certificates store
            if str2bool(getenv("QDT_SSL_USE_SYSTEM_STORES", False)):
                logger.debug(
                    "Option to use native system certificates stores is enabled."
                )
                dl_session.mount("https://", TruststoreAdapter())

            with dl_session.get(
                url=requote_uri(remote_url_to_download),
                stream=use_stream,
                timeout=timeout,
            ) as req:
                req.raise_for_status()
                if use_stream:
                    with local_file_path.open(mode="wb") as buffile:
                        for chunk in req.iter_content(chunk_size=chunk_size):
                            if chunk:
                                buffile.write(chunk)
                else:
                    # Download download the entire content at once
                    local_file_path.write_bytes(req.content)

            logger.info(
                f"Downloading {remote_url_to_download} to {local_file_path} "
                f"({convert_octets(local_file_path.stat().st_size)}) succeeded."
            )
    except HTTPError as error:
        logger.error(
            f"Downloading {remote_url_to_download} to {local_file_path} failed. "
            f"Cause: HTTPError. Trace: {error}."
        )
        if isinstance(req, Response):
            http_error_details = {
                "status": req.status_code,
                "headers": req.headers,
                "body": req.content,
            }
            logger.error(
                f"Addtional details grabbed from HTTP response: {http_error_details}"
            )

        raise error
    except ConnectionError as error:
        logger.error(
            f"Downloading {remote_url_to_download} to {local_file_path} failed. "
            f"Cause: ConnectionError. Trace: {error}"
        )
        raise error
    except Exception as error:
        logger.error(
            f"Downloading {remote_url_to_download} to {local_file_path} failed. "
            f"Cause: Unknown error. Trace: {error}",
            stack_info=True,
        )
        raise error

    return local_file_path



#! python3  # noqa: E265

"""
    Rules context.

    Author: Julien Moura (https://github.com/guts)
"""


# #############################################################################
# ########## Libraries #############
# ##################################

# Standard library
import json
import logging
import platform
from datetime import date
from getpass import getuser
from sys import platform as opersys

# package
from qgis_deployment_toolbelt.utils.user_groups import (
    get_user_domain_groups,
    get_user_local_groups,
)
from qgis_deployment_toolbelt.utils.win32utils import (
    ExtendedNameFormat,
    get_current_user_extended_data,
)

# #############################################################################
# ########## Globals ###############
# ##################################

# logs
logger = logging.getLogger(__name__)


# #############################################################################
# ########## Functions #############
# ##################################


class QdtRulesContext:

    @property
    def _context_date(self) -> dict:
        """Returns a context dictionary with date informations that can be used in QDT
        various places: rules...

        Returns:
            dict: dict with current date informations
        """
        today = date.today()
        return {
            "current_day": today.day,
            "current_weekday": today.weekday(),  # monday = 0, sunday = 6
            "current_month": today.month,
            "current_year": today.year,
        }

    @property
    def _context_environment(self) -> dict:
        """Returns a dictionary containing some environment information (computer, network,
            platform) that can be used in QDT various places: rules...

        Returns:
            dict: dict with some environment metadata to use in rules.
        """
        try:
            linux_distribution_name = f"{platform.freedesktop_os_release().get('NAME')}"
            linux_distribution_version = (
                f"{platform.freedesktop_os_release().get('VERSION_ID')}"
            )
        except OSError as err:
            logger.debug(
                f"Unable to determine current Linux distribution. Trace: {err}."
            )
            linux_distribution_name = None
            linux_distribution_version = None

        return {
            "computer_network_name": platform.node(),
            "operating_system_code": opersys,
            "operating_system_release": platform.release(),
            "processor_architecture": platform.machine(),
            # custom Linux
            "linux_distribution_name": linux_distribution_name,
            "linux_distribution_version": linux_distribution_version,
            # custom Windows
            "windows_edition": platform.win32_edition(),
        }

    @property
    def _context_user(self) -> dict:
        """Returns a dictionary containing user informations that can be used in QDT Rules
            context.

        Returns:
            dict: dict user information.
        """
        if opersys == "win32":
            windows_extended = {
                k.name: get_current_user_extended_data(k) for k in ExtendedNameFormat
            }
        else:
            windows_extended = None

        try:
            user_domain_groups = get_user_domain_groups()
        except Exception as err:
            logger.error(f"Unable to retrieve user domain groups. Trace: {err}")
            user_domain_groups = []

        return {
            "name": getuser(),
            "groups_local": get_user_local_groups(),
            "groups_domain": user_domain_groups,
            "windows_extended": windows_extended,
        }

    # -- EXPORT
    def to_dict(self) -> dict:
        """Convert object into dictionary.

        Returns:
            dict: object as dictionary
        """
        result = {}
        for attr in dir(self):
            if isinstance(
                getattr(self.__class__, attr, None), property
            ) and attr.startswith("_context_"):
                result[attr.removeprefix("_context_")] = getattr(self, attr)
        return result

    def to_json(self, **kwargs) -> str:
        """Supersedes json.dumps using the dictionary returned by to_dict().
        kwargs are passed to json.dumps.

        Returns:
            str: object serialized as JSON string

        Example:

            .. code-block:: python

                from pathlib import Path

                rules_context = QdtRulesContext()

                # write into the file passing extra parameters to json.dumps
                with Path("qdt_rules_context.json").open("w", encoding="UTF8") as wf:
                    wf.write(rules_context.to_json(indent=4, sort_keys=True))
        """
        obj_as_dict = self.to_dict()

        return json.dumps(obj_as_dict, **kwargs)



#! python3  # noqa E265

"""
    Usage from the repo root folder:

    .. code-block:: bash
        # for whole tests
        python -m unittest tests.test_utils
        # for specific test
        python -m unittest tests.test_utils.TestUtilsLinux.test_win32_getenv
"""

# standard library
import tempfile
import unittest
from pathlib import Path
from sys import platform as opersys

# project
from qgis_deployment_toolbelt.utils import linux_utils

# ############################################################################
# ########## Classes #############
# ################################


class TestUtilsLinux(unittest.TestCase):
    """Test package utilities."""

    @unittest.skipIf(opersys != "linux", "Test specific to Linux.")
    def test_file_to_use_for_storing_environment_variables(self):
        """Test file priority in bash context."""
        with tempfile.TemporaryDirectory(
            prefix="qdt_test_linux_utils_bash_profile"
        ) as tmp_dir_name:
            # reference fake bash files
            tmp_fake_dot_profile = Path(tmp_dir_name).joinpath(".profile")
            linux_utils.bash_user_dot_profile = tmp_fake_dot_profile
            tmp_fake_dot_bash_login = Path(tmp_dir_name).joinpath(".bash_login")
            linux_utils.bash_user_dot_bash_login = tmp_fake_dot_bash_login
            tmp_fake_dot_bash_profile = Path(tmp_dir_name).joinpath(".bash_profile")
            linux_utils.bash_user_dot_bash_profile = tmp_fake_dot_bash_profile

            # create fake .profile
            tmp_fake_dot_profile.touch()
            self.assertEqual(
                linux_utils.bash_user_which_file_to_store(),
                tmp_fake_dot_profile,
            )

            # create fake .bash_login
            tmp_fake_dot_bash_login.touch()
            self.assertEqual(
                linux_utils.bash_user_which_file_to_store(),
                tmp_fake_dot_bash_login,
            )

            # create fake bash_profile
            tmp_fake_dot_bash_profile.touch()
            linux_utils.bash_user_dot_bash_profile = Path(tmp_dir_name).joinpath(
                ".bash_profile"
            )
            self.assertEqual(
                linux_utils.bash_user_which_file_to_store(),
                tmp_fake_dot_bash_profile,
            )


# ############################################################################
# ####### Stand-alone run ########
# ################################
if __name__ == "__main__":
    unittest.main()



#! python3  # noqa: E265

"""
    Tools to find installed QGIS version

    Author: Jean-Marie KERLOCH (https://github.com/jmkerloch)
"""

# #############################################################################
# ########## Libraries #############
# ##################################


# Standard library
import logging
import os
import re
import subprocess
from os.path import expandvars
from pathlib import Path
from shutil import which
from sys import platform as opersys

# package
from qgis_deployment_toolbelt.exceptions import QgisInstallNotFound
from qgis_deployment_toolbelt.jobs.generic_job import GenericJob

# #############################################################################
# ########## Globals ###############
# ##################################


# logs
logger = logging.getLogger(__name__)

# #############################################################################
# ########## Classes ###############
# ##################################


class JobQgisInstallationFinder(GenericJob):
    """
    Class to find installed QGIS version
    """

    ID: str = "qgis-installation-finder"
    OPTIONS_SCHEMA: dict = {
        "version_priority": {
            "type": (list, str),
            "required": False,
            "default": None,
            "possible_values": None,
            "condition": None,
        },
        "if_not_found": {
            "type": str,
            "required": False,
            "default": "warning",
            "possible_values": ("warning", "error"),
            "condition": "in",
        },
    }

    def __init__(self, options: dict) -> None:
        """Instantiate the class.
        Args:
            options (dict): job options
        """

        super().__init__()
        self.options: dict = self.validate_options(options)

    def run(self) -> None:
        """Define QDT_QGIS_EXE_PATH for shortcut creation"""
        if not self.run_needed():
            return

        if opersys not in ("linux", "win32"):
            logger.error(f"This job does not support your operating system: {opersys}")
            return

        installed_qgis_path: str | None = self.get_installed_qgis_path()

        if installed_qgis_path:
            os.environ["QDT_QGIS_EXE_PATH"] = installed_qgis_path
        else:
            if self.options.get("if_not_found", "warning") == "error":
                raise QgisInstallNotFound()
            else:
                logger.warning("No QGIS installation found")

        # log
        logger.debug(f"Job {self.ID} ran successfully.")

    # -- INTERNAL LOGIC ------------------------------------------------------
    def run_needed(self) -> bool:
        """Check if job run is needed

        Returns:
            bool: return True if job must be run, False otherwise
        """
        if "QDT_QGIS_EXE_PATH" in os.environ:
            qgis_bin = os.environ["QDT_QGIS_EXE_PATH"]
            if Path(qgis_bin).exists():
                version_str = self._get_qgis_bin_version(qgis_bin)
                if version_str:
                    logger.info(
                        f"QDT_QGIS_EXE_PATH defined and path {qgis_bin} exists for QGIS {version_str}. {self.ID} job is skipped. "
                    )
                    return False
                else:
                    logger.warning(
                        f"QDT_QGIS_EXE_PATH defined and path {qgis_bin} exists but the QGIS version can't be defined. Check variable."
                    )
        return True

    def get_installed_qgis_path(self) -> str | None:
        """Get list of installed qgis

        Returns:
            str | None : installed qgis path
        """
        if opersys == "linux":
            found_versions = self._get_linux_installed_qgis_path()
        elif opersys == "win32":
            # Check for installed version in the default install directory
            found_versions = self._get_windows_installed_qgis_path()

        if len(found_versions) == 0:
            return None

        logger.debug(f"Found installed QGIS : {found_versions}")
        latest_version = self._get_latest_version_from_list(
            versions=list(found_versions.keys())
        )
        latest_qgis = found_versions[latest_version]

        version_priority: list[str] = []
        if "version_priority" in self.options:
            version_priority = self.options["version_priority"]

        # Add preferred qgis version on top of the list
        if "QDT_PREFERRED_QGIS_VERSION" in os.environ:
            version_priority.insert(0, os.environ["QDT_PREFERRED_QGIS_VERSION"])

        for version in version_priority:
            if latest_matching_version := self._get_latest_matching_version_path(
                found_versions=found_versions, version=version
            ):
                return latest_matching_version

            version_priority_str = ",".join(self.options["version_priority"])
            logger.info(
                f"QGIS version(s) [{version_priority_str}] not found. Using most recent found version {latest_version} : {latest_qgis}"
            )

        return latest_qgis

    @staticmethod
    def _get_latest_version_from_list(versions: list[str]) -> str | None:
        """Get latest version from a list, OSGEO4W are last

        Args:
            versions (list[str]): list of found version

        Returns:
            str | None: latest version, None if no version provided
        """
        if len(versions):
            used_version = versions
            used_version.sort(reverse=True)
            return used_version[0]
        return None

    @staticmethod
    def _get_latest_matching_version_path(
        found_versions: dict[str, str], version: str
    ) -> str | None:
        """Get latest version path matching a wanted version

        Args:
            found_versions (dict[str, str]): dict of found versions
            version (str): wanted version

        Returns:
            str | None: latest matching version path, None if not found
        """
        match_versions = []
        for found_version in found_versions.keys():
            if found_version.startswith(version):
                match_versions.append(found_version)
        # Use latest version
        if latest_version := JobQgisInstallationFinder._get_latest_version_from_list(
            match_versions
        ):
            return found_versions[latest_version]
        return None

    @staticmethod
    def _get_qgis_bin_in_install_dir(install_dir: str) -> str | None:
        """Get QGIS binary path from an install directory

        Args:
            install_dir (str): install directory

        Returns:
            str | None: QGIS bin path, None if not found
        """
        binary_pattern = re.compile(r"qgis(-ltr)?-bin\.exe", re.IGNORECASE)
        # Check if the bin directory exists within this directory
        bin_dir = os.path.join(install_dir, "bin")
        if os.path.exists(bin_dir):
            # Check if any binary file matches the pattern in the bin directory
            for filename in os.listdir(bin_dir):
                if binary_pattern.match(filename):
                    qgis_exe = os.path.join(bin_dir, filename)
                    return qgis_exe
        return None

    @staticmethod
    def _get_windows_installed_qgis_path() -> dict[str, str]:
        """Get dict of installed QGIS version in common install directory

        Returns:
            dict[str, str]: dict of QGIS version and QGIS bin path
        """
        # Check for installed version in the default install directory
        found_version = {}

        # Program files
        prog_file_dir = expandvars("%PROGRAMFILES%")
        directory_pattern = re.compile(r"QGIS (\d+)\.(\d+)\.(\d+)", re.IGNORECASE)
        for dir_name in os.listdir(prog_file_dir):
            # Check if the directory name matches the pattern
            match = directory_pattern.match(dir_name)
            if match:
                install_dir = os.path.join(prog_file_dir, dir_name)
                if qgis_bin := JobQgisInstallationFinder._get_qgis_bin_in_install_dir(
                    install_dir
                ):
                    version_str = JobQgisInstallationFinder._get_qgis_bin_version(
                        qgis_bin=qgis_bin
                    )
                    if version_str:
                        found_version[version_str] = qgis_bin
                    else:
                        logger.warning(
                            f"Can't define QGIS version for '{qgis_bin}' binary."
                        )

        # OSGEO4W
        install_dir = os.environ.get("QDT_OSGEO4W_INSTALL_DIR", "C:\\OSGeo4W")
        if qgis_bin := JobQgisInstallationFinder._get_qgis_bin_in_install_dir(
            install_dir
        ):
            version_str = JobQgisInstallationFinder._get_qgis_bin_version(
                qgis_bin=qgis_bin
            )
            if version_str:
                found_version[version_str] = qgis_bin
            else:
                logger.warning(f"Can't define QGIS version for '{qgis_bin}' binary.")

        return found_version

    @staticmethod
    def _get_linux_installed_qgis_path() -> dict[str, str]:
        """Get install qgis path for linux operating system with which

        Returns:
            dict[str, str]: dict of QGIS version and QGIS bin path
        """
        # use which to find installed qgis
        found_version = {}
        if qgis_bin := which("qgis"):
            logger.debug(f"QGIS path found using which: {qgis_bin}")
            version_str = JobQgisInstallationFinder._get_qgis_bin_version(
                qgis_bin=qgis_bin
            )
            if version_str:
                found_version[version_str] = qgis_bin
            else:
                logger.warning(f"Can't define QGIS version for '{qgis_bin}' binary.")

        return found_version

    @staticmethod
    def _get_qgis_bin_version(qgis_bin: str) -> str | None:
        """Get QGIS bin version with --version

        Args:
            qgis_bin (str): QGIS bin path

        Returns:
            str | None: SemVer version, None if version not found
        """
        process = subprocess.Popen(
            f'"{qgis_bin}" --version', stdout=subprocess.PIPE, shell=True
        )
        stdout_, _ = process.communicate()
        version_str = stdout_.decode()
        version_pattern = r"QGIS (\d+\.\d+\.\d+)-(\w+).*"
        version_match = re.match(version_pattern, version_str)
        if version_match:
            return version_match.group(1)
        return None



#! python3  # noqa E265

"""Usage from the repo root folder:

    .. code-block:: python

        # for whole test
        python -m unittest tests.test_jobs_launch
        # for specific
        python -m unittest tests.test_jobs_launch.TestJobsLaunch.test_jobs_launcher
"""

# #############################################################################
# ########## Libraries #############
# ##################################

# Standard library
import unittest
from pathlib import Path

# package
from qgis_deployment_toolbelt.jobs import JobsOrchestrator
from qgis_deployment_toolbelt.scenarios.scenario_reader import ScenarioReader

# #############################################################################
# ########## Classes ###############
# ##################################


class TestJobsLaunch(unittest.TestCase):
    """Test module."""

    # -- Standard methods --------------------------------------------------------
    @classmethod
    def setUpClass(cls):
        """Executed when module is loaded before any test."""
        cls.good_scenario_files = sorted(
            Path("tests/fixtures/").glob("scenarios/good_*.y*ml")
        )

    # -- TESTS ---------------------------------------------------------
    def test_jobs_launcher(self):
        """Test YAML loader"""
        steps_ok = []
        orchestrator = JobsOrchestrator()

        for i in self.good_scenario_files:
            scenario = ScenarioReader(in_yaml=i)
            for step in scenario.steps:
                if step.get("uses") not in orchestrator.jobs_ids:
                    continue
                else:
                    steps_ok.append(step)

        for step in steps_ok:
            job = orchestrator.init_job_class_from_id(
                job_id=step.get("uses"), options=step.get("with")
            )
            self.assertIsNotNone(job)



#! python3  # noqa E265

"""
    Usage from the repo root folder:

    .. code-block:: bash
        # for whole tests
        python -m unittest tests.test_utils_ini_interpolation
        # for specific test
        python -m unittest tests.test_utils_ini_interpolation.TestUtilsIniCustomInterpolation.test_ini_env_var_interpolation
"""

# standard library
import tempfile
import unittest
from configparser import ConfigParser
from getpass import getuser
from os import environ, getenv
from pathlib import Path
from sys import platform as opersys

# project
from qgis_deployment_toolbelt.utils.ini_interpolation import (
    EnvironmentVariablesInterpolation,
)

# ############################################################################
# ########## Classes #############
# ################################


class TestUtilsIniCustomInterpolation(unittest.TestCase):
    """Test custom INI values' interpolation."""

    def check_config_read_linux(self, config: ConfigParser) -> None:
        """Check config with environnement variable interpolation for linux

        Args:
            config (ConfigParser): parser to check
        """
        self.assertEqual(config.get(section="test", option="user"), getuser())
        self.assertEqual(config.get(section="test", option="user_home"), getenv("HOME"))
        self.assertEqual(
            config.get(section="test", option="fake_value_from_environment_variable"),
            getenv("QDT_TEST_ENV_VARIABLE"),
        )
        self.assertEqual(
            config.get(section="test", option="qdt_working_directory"),
            f"{config.get(section='test', option='user_home')}/.cache/qgis-deployment-toolbelt",
        )

    @unittest.skipIf(opersys == "win32", "Test specific to Linux.")
    def test_ini_env_var_interpolation_linux(self):
        """Test interpolation results."""
        fixtures_configuration = """
        [test]
        user=$USER
        user_home = $HOME
        fake_value_from_environment_variable = $QDT_TEST_ENV_VARIABLE
        qdt_working_directory = %(user_home)s/.cache/qgis-deployment-toolbelt

        """
        environ["QDT_TEST_ENV_VARIABLE"] = "TEST_VALUE"

        config = ConfigParser(interpolation=EnvironmentVariablesInterpolation())
        config.read_string(fixtures_configuration)
        self.check_config_read_linux(config=config)

        environ.pop("QDT_TEST_ENV_VARIABLE")

    @unittest.skipIf(opersys == "win32", "Test specific to Linux.")
    def test_ini_env_var_interpolation_write_linux(self):
        """Test interpolation results for ini write."""
        fixtures_configuration = """
        [test]
        user=$USER
        user_home = $HOME
        fake_value_from_environment_variable = $QDT_TEST_ENV_VARIABLE
        qdt_working_directory = %(user_home)s/.cache/qgis-deployment-toolbelt

        """
        environ["QDT_TEST_ENV_VARIABLE"] = "TEST_VALUE"

        config = ConfigParser(interpolation=EnvironmentVariablesInterpolation())
        config.read_string(fixtures_configuration)

        with tempfile.TemporaryDirectory(
            prefix="qdt_test_write_env_ini_file_", ignore_cleanup_errors=True
        ) as tmpdirname:
            # Write new .ini : environnement variable will be written
            tmp_copy = Path(tmpdirname).joinpath("new_file.ini")
            with open(tmp_copy, "w") as f:
                config.write(f)
            # Check that environnement variable are replaced
            default_config_parser = ConfigParser()
            default_config_parser.read(tmp_copy)
            self.check_config_read_linux(config=default_config_parser)

        environ.pop("QDT_TEST_ENV_VARIABLE")

    def check_config_read_windows(self, config: ConfigParser) -> None:
        """Check config with environnement variable interpolation for windows

        Args:
            config (ConfigParser): parser to check
        """
        self.assertEqual(config.get(section="test", option="user"), getuser())
        self.assertEqual(
            config.get(section="test", option="user_home"), getenv("USERPROFILE")
        )
        self.assertEqual(
            config.get(section="test", option="fake_value_from_environment_variable"),
            getenv("QDT_TEST_ENV_VARIABLE"),
        )
        self.assertEqual(
            config.get(section="test", option="qdt_working_directory"),
            f"{config.get(section='test', option='user_home')}/.cache/qgis-deployment-toolbelt",
        )

    @unittest.skipIf(opersys != "win32", "Test specific to Windows.")
    def test_ini_env_var_interpolation_windows(self):
        """Test interpolation results."""
        fixtures_configuration = """
        [test]
        user = $USERNAME
        user_home = $USERPROFILE
        fake_value_from_environment_variable = $QDT_TEST_ENV_VARIABLE
        qdt_working_directory = %(user_home)s/.cache/qgis-deployment-toolbelt

        """
        environ["QDT_TEST_ENV_VARIABLE"] = "TEST_VALUE"

        config = ConfigParser(interpolation=EnvironmentVariablesInterpolation())
        config.read_string(fixtures_configuration)
        self.check_config_read_windows(config=config)

        environ.pop("QDT_TEST_ENV_VARIABLE")

    @unittest.skipIf(opersys != "win32", "Test specific to Windows.")
    def test_ini_env_var_interpolation_write_windows(self):
        """Test interpolation results."""
        fixtures_configuration = """
        [test]
        user = $USERNAME
        user_home = $USERPROFILE
        fake_value_from_environment_variable = $QDT_TEST_ENV_VARIABLE
        qdt_working_directory = %(user_home)s/.cache/qgis-deployment-toolbelt

        """
        environ["QDT_TEST_ENV_VARIABLE"] = "TEST_VALUE"

        config = ConfigParser(interpolation=EnvironmentVariablesInterpolation())
        config.read_string(fixtures_configuration)
        with tempfile.TemporaryDirectory(
            prefix="qdt_test_write_env_ini_file_", ignore_cleanup_errors=True
        ) as tmpdirname:
            # Write new .ini : environnement variable will be written
            tmp_copy = Path(tmpdirname).joinpath("new_file.ini")
            with open(tmp_copy, "w") as f:
                config.write(f)
            # Check that environnement variable are replaced
            default_config_parser = ConfigParser()
            default_config_parser.read(tmp_copy)
            self.check_config_read_windows(config=default_config_parser)

        environ.pop("QDT_TEST_ENV_VARIABLE")


# ############################################################################
# ####### Stand-alone run ########
# ################################
if __name__ == "__main__":
    unittest.main()



#! python3  # noqa: E265


"""
    Sub-command to export local rules context.

    Author: Julien M. (https://github.com/guts)
"""


# ############################################################################
# ########## IMPORTS #############
# ################################

# standard library
import argparse
import logging
from pathlib import Path

# package
from qgis_deployment_toolbelt.constants import get_qdt_working_directory
from qgis_deployment_toolbelt.profiles.rules_context import QdtRulesContext
from qgis_deployment_toolbelt.utils.bouncer import exit_cli_error, exit_cli_success

# ############################################################################
# ########## GLOBALS #############
# ################################

logger = logging.getLogger(__name__)


# ############################################################################
# ########## CLI #################
# ################################


def parser_rules_context_export(
    subparser: argparse.ArgumentParser,
) -> argparse.ArgumentParser:
    """Set the argument parser for subcommand.

    Args:
        subparser (argparse.ArgumentParser): parser to set up

    Returns:
        argparse.ArgumentParser: parser ready to use
    """

    subparser.add_argument(
        "-o",
        "--output",
        help="Path to the output file where to write rules context.",
        default=get_qdt_working_directory().joinpath("export/qdt_rules_context.json"),
        type=Path,
        dest="output_path",
    )

    subparser.set_defaults(func=run)

    return subparser


# ############################################################################
# ########## MAIN ################
# ################################


def run(args: argparse.Namespace):
    """Run the sub command logic.

    Open result of a previous command.

    Args:
        args (argparse.Namespace): arguments passed to the subcommand
    """
    logger.debug(f"Running {args.command} with {args}")

    try:
        context_json_path = Path(args.output_path)
        context_json_path.parent.mkdir(parents=True, exist_ok=True)
        rules_context = QdtRulesContext()

        # write into the file passing extra parameters to json.dumps
        with context_json_path.open("w", encoding="UTF8") as wf:
            wf.write(rules_context.to_json(indent=4, sort_keys=True))

        # exit nicely
        print(f"Rules context exported in {args.output_path}")
        exit_cli_success(f"Rules context exported in {args.output_path}")
    except Exception as err:
        exit_cli_error(err)



#! python3  # noqa: E265

"""
    Manage splash screens for QGIS Profiles.

    Author: Julien Moura (https://github.com/guts)
"""


# #############################################################################
# ########## Libraries #############
# ##################################

# Standard library
import logging
from pathlib import Path

# package
from qgis_deployment_toolbelt.exceptions import SplashScreenBadDimensions
from qgis_deployment_toolbelt.jobs.generic_job import GenericJob
from qgis_deployment_toolbelt.profiles.qdt_profile import QdtProfile
from qgis_deployment_toolbelt.profiles.qgis_ini_handler import QgisIniHelper
from qgis_deployment_toolbelt.utils.check_image_size import check_image_dimensions

# #############################################################################
# ########## Globals ###############
# ##################################

# logs
logger = logging.getLogger(__name__)


# #############################################################################
# ########## Classes ###############
# ##################################


class JobSplashScreenManager(GenericJob):
    """
    Job to set-up splash screen for QGIS profile.
    """

    ID: str = "splash-screen-manager"
    OPTIONS_SCHEMA: dict = {
        "action": {
            "type": str,
            "required": False,
            "default": "create_or_restore",
            "possible_values": ("create", "create_or_restore", "remove"),
            "condition": "in",
        },
        "strict": {
            "type": bool,
            "required": False,
            "default": False,
            "possible_values": None,
            "condition": None,
        },
    }
    DEFAULT_SPLASH_FILEPATH: str = "images/splash.png"
    SPLASH_FILENAME: str = "splash.png"

    def __init__(self, options: dict) -> None:
        """Instantiate the class.

        Args:
            options (dict): dictionary of options.
        """
        super().__init__()
        self.options: dict = self.validate_options(options)

    def run(self) -> None:
        """Execute job logic."""
        # check of there are some profiles folders within the downloaded folder
        downloaded_profiles = self.list_downloaded_profiles()

        if downloaded_profiles is None:
            logger.error("No QGIS profile found in the downloaded folder.")
            return

        # iterate over downloaded profiles
        for profile_downloaded in downloaded_profiles:
            logger.debug(
                f"Applying splash screen for downloaded profile '{profile_downloaded.name}'"
            )
            profile_installed = profile_downloaded.installed_profile
            # target QGIS configuration files
            if isinstance(profile_installed, QdtProfile):
                qini_helper_installed = profile_installed.get_qgis3ini_helper()
            else:
                qini_helper_installed = QgisIniHelper(
                    ini_filepath=profile_downloaded.path_in_qgis / "QGIS/QGIS3.ini",
                    ini_type="profile_qgis3",
                )

            if self.options.get("action") == "remove":
                qini_helper_installed.set_splash_screen(switch=False)
            elif self.options.get("action") in ("create", "create_or_restore"):
                # default absolute splash screen path
                installed_splash_screen_filepath = (
                    profile_downloaded.path_in_qgis / self.DEFAULT_SPLASH_FILEPATH
                )
                # check if a profile.json exists
                if isinstance(profile_installed, QdtProfile):
                    # if the splash image referenced into the profile.json exists, make
                    # sure it complies QGIS splash screen naming rules
                    if (
                        isinstance(profile_installed.splash, Path)
                        and profile_installed.splash.is_file()
                    ):
                        # make sure that the filename complies with what QGIS expects
                        if (
                            profile_installed.splash.name
                            != installed_splash_screen_filepath.name
                        ):
                            splash_filepath = profile_installed.splash.with_name(
                                self.SPLASH_FILENAME
                            )
                            profile_installed.splash.replace(splash_filepath)
                            logger.debug(
                                f"Specified splash screen renamed into {splash_filepath} "
                                "to comply with QGIS expectations."
                            )
                        else:
                            # homogeneize filepath var name
                            logger.debug(
                                f"{profile_installed.name} Splash screen already exists "
                                f"at {profile_installed.splash} and it complies QGIS "
                                "rules."
                            )

                else:
                    logger.info(
                        f"No profile.json found for profile '{profile_downloaded.folder}'"
                    )

                # now, splash screen image should be at {profile_dir}/images/splash.png
                if not installed_splash_screen_filepath.is_file():
                    if isinstance(profile_installed, QdtProfile):
                        logger.debug(
                            "No splash screen found or defined for profile: "
                            f"{profile_installed.name} ({profile_installed.path_in_qgis})"
                        )
                    else:
                        logger.debug(
                            "No splash screen found or defined for profile: "
                            f"{profile_downloaded.folder} ({profile_downloaded.path_in_qgis})"
                        )
                    continue

                # check image size to fit QGIS restrictions
                is_img_compliant = check_image_dimensions(
                    image_filepath=installed_splash_screen_filepath.resolve(),
                    max_width=605,
                    max_height=305,
                    allowed_images_extensions=(".png",),
                )
                if not is_img_compliant:
                    err = SplashScreenBadDimensions(
                        image_filepath=installed_splash_screen_filepath,
                        profile_name=profile_installed.name,
                    )
                    if self.options.get("strict") is True:
                        raise err
                    else:
                        logger.warning(err.message)

                # enable UI customization
                qini_helper_installed.set_ui_customization_enabled(
                    switch=True,
                )
                logger.info(
                    f"Profile {profile_installed.name}: customization enabled "
                    f"in {qini_helper_installed.profile_config_path}"
                )

                # set the splash screen into the customization file
                qini_helper_installed.set_splash_screen(
                    ini_file=qini_helper_installed.profile_customization_path,
                    splash_screen_filepath=installed_splash_screen_filepath.resolve(),
                    switch=True,
                )
                logger.info(
                    f"Profile {profile_installed.name}: splash screen set "
                    f"in {qini_helper_installed.profile_customization_path}"
                )

            else:
                raise NotImplementedError

        logger.debug(f"Job {self.ID} ran successfully.")



#! python3  # noqa: E265

"""
    Base of QDT jobs.

    Author: Julien Moura (https://github.com/guts)
"""


# #############################################################################
# ########## Libraries #############
# ##################################

# Standard library
import logging
from functools import lru_cache
from os import getenv
from pathlib import Path

# 3rd party
from python_rule_engine import RuleEngine

# package
from qgis_deployment_toolbelt.constants import (
    OSConfiguration,
    get_qdt_working_directory,
)
from qgis_deployment_toolbelt.exceptions import (
    JobOptionBadName,
    JobOptionBadValue,
    JobOptionBadValueType,
)
from qgis_deployment_toolbelt.profiles.qdt_profile import QdtProfile
from qgis_deployment_toolbelt.profiles.rules_context import QdtRulesContext

# #############################################################################
# ########## Globals ###############
# ##################################

# logs
logger = logging.getLogger(__name__)

# #############################################################################
# ########## Classes ###############
# ##################################


class GenericJob:
    """Generic base for QDT jobs."""

    ID: str = ""
    OPTIONS_SCHEMA: dict[dict] = dict(dict())

    def __init__(self) -> None:
        """Object instanciation."""
        # operating system configuration
        self.os_config = OSConfiguration.from_opersys()
        self.qdt_rules_context = QdtRulesContext()

        # local QDT folders
        self.qdt_working_folder = get_qdt_working_directory()
        if not self.qdt_working_folder.exists():
            logger.info(
                f"QDT downloaded folder not found: {self.qdt_working_folder}. "
                "Creating it to properly run the job."
            )
            self.qdt_working_folder.mkdir(parents=True, exist_ok=True)
        logger.debug(f"QDT working folder: {self.qdt_working_folder}")

        self.qdt_downloaded_repositories = self.qdt_working_folder.joinpath(
            f"repositories/{getenv('QDT_TMP_RUNNING_SCENARIO_ID', 'default')}"
        )
        self.qdt_plugins_folder = self.qdt_working_folder.joinpath("plugins")

        # destination profiles folder
        self.qgis_profiles_path: Path = self.os_config.qgis_profiles_path
        if not self.qgis_profiles_path.exists():
            logger.info(
                f"Installed QGIS profiles folder not found: {self.qgis_profiles_path}. "
                "Creating it to properly run the job."
            )
            self.qgis_profiles_path.mkdir(parents=True)
        logger.debug(f"Installed QGIS profiles folder: {self.qgis_profiles_path}")

    def list_downloaded_profiles(self) -> tuple[QdtProfile] | None:
        """List downloaded QGIS profiles, i.e. a profile's folder located into the QDT
            working folder.
            Typically: `~/.cache/qgis-deployment-toolbelt/repositories/geotribu` or
            `%USERPROFILE%/.cache/qgis-deployment-toolbelt/repositories/geotribu`).

        Returns:
            tuple[QdtProfile] | None: tuple of profiles objects or None if no profile
                folder listed
        """
        return self.filter_profiles_folder(
            start_parent_folder=self.qdt_downloaded_repositories
        )

    def list_installed_profiles(self) -> tuple[QdtProfile] | None:
        """List installed QGIS profiles, i.e. a profile's folder located into the QGIS
            profiles path and so accessible to the end-user through the QGIS interface.
            Typically: `~/.local/share/QGIS/QGIS3/profiles/geotribu` or
            `%APPDATA%/QGIS/QGIS3/profiles/geotribu`).

        Returns:
            tuple[QdtProfile] | None: tuple of profiles objects or Non if no profile is
                installed in QGIS3/profiles
        """
        return self.filter_profiles_folder(start_parent_folder=self.qgis_profiles_path)

    def filter_profiles_folder(
        self, start_parent_folder: Path
    ) -> tuple[QdtProfile, ...] | None:
        """Parse a folder structure to filter on QGIS profiles folders.

        Returns:
            tuple[QdtProfile] | None: tuple of profiles objects matching criteria or
                None if no profile folder found
        """
        # first, try to get folders containing a profile.json
        li_qgis_qdt_profiles: list[QdtProfile] = [
            QdtProfile.from_json(profile_json_path=f, profile_folder=f.parent)
            for f in start_parent_folder.glob("**/profile.json")
        ]

        if not len(li_qgis_qdt_profiles):
            logger.error(f"No QGIS profile found in {start_parent_folder}.")
            return

        logger.debug(
            f"{len(li_qgis_qdt_profiles)} profiles found within {start_parent_folder}"
        )

        # filter out profiles that do not match the rules
        profiles_matched, profiles_unmatched = self.filter_profiles_on_rules(
            tup_qdt_profiles=tuple(li_qgis_qdt_profiles)
        )

        if not len(profiles_matched):
            logger.warning(
                f"None of the {len(li_qgis_qdt_profiles)} profiles meet the deployment "
                "requirements."
            )
            return

        if len(profiles_unmatched):
            logger.info(
                f"{len(profiles_unmatched)}/{len(li_qgis_qdt_profiles)} profiles "
                "do not meet the conditions for deployment: "
                f"{', '.join([p.name for p in profiles_unmatched])}"
            )

        return tuple(profiles_matched)

    @lru_cache(maxsize=1024)
    def filter_profiles_on_rules(
        self, tup_qdt_profiles: tuple[QdtProfile]
    ) -> tuple[list[QdtProfile], list[QdtProfile]]:
        """Evaluate profile regarding to its deployment rules.

        Args:
            tup_qdt_profiles (tuple[QdtProfile]): input tuple of QDT profiles

        Returns:
            tuple[list[QdtProfile], list[QdtProfile]]: tuple of profiles that matched
            and those which did not match their deployment rules
        """
        li_profiles_matched = []
        li_profiles_unmatched = []

        for profile in tup_qdt_profiles:
            if profile.rules is None:
                logger.debug(f"No rules to apply to {profile.name}")
                li_profiles_matched.append(profile)
                continue

            logger.debug(
                f"Checking that profile '{profile.name}' matches deployment conditions. "
                f"{len(profile.rules)} rules found."
            )
            try:
                engine = RuleEngine(rules=profile.rules)
                results = engine.evaluate(obj=self.qdt_rules_context.to_dict())
                if len(results) == len(profile.rules):
                    logger.debug(
                        f"Profile '{profile.name}' matches {len(profile.rules)} "
                        "deployment rule(s)."
                    )
                    li_profiles_matched.append(profile)
                else:
                    logger.info(
                        f"Profile '{profile.name}' does not match the deployment "
                        f"conditions: {len(results)}/{len(profile.rules)} rule(s) "
                        "matched."
                    )
                    li_profiles_unmatched.append(profile)

            except Exception as err:
                logger.error(
                    f"Error occurred parsing rules of profile '{profile.name}'. "
                    f"Trace: {err}"
                )

        return li_profiles_matched, li_profiles_unmatched

    def validate_options(self, options: dict[dict]) -> dict[dict]:
        """Validate options.

        Args:
            options (dict[dict]): options to validate.

        Raises:
            ValueError: if option has an invalid name or doesn't comply with condition
            TypeError: if the option does'nt not comply with expected type

        Returns:
            dict[dict]: options if valid
        """
        if not isinstance(options, dict):
            raise TypeError(f"Options to validate must be a dict, not {type(options)}.")

        for option in options:
            if option not in self.OPTIONS_SCHEMA:
                raise JobOptionBadName(
                    job_id=self.ID,
                    bad_option_name=option,
                    expected_options_names=self.OPTIONS_SCHEMA.keys(),
                )

            option_in = options.get(option)
            option_def: dict = self.OPTIONS_SCHEMA.get(option)
            # check value type
            if not isinstance(option_in, option_def.get("type")):
                raise JobOptionBadValueType(
                    job_id=self.ID,
                    bad_option_name=option,
                    bad_option_value=option_in,
                    expected_option_type=option_def.get("type"),
                )
            # check value condition
            if option_def.get("condition") == "startswith" and not option_in.startswith(
                option_def.get("possible_values")
            ):
                raise JobOptionBadValue(
                    job_id=self.ID,
                    bad_option_name=option,
                    bad_option_value=option_in,
                    condition="startswith",
                    accepted_values=option_def.get("possible_values"),
                )
            elif option_def.get(
                "condition"
            ) == "in" and option_in not in option_def.get("possible_values"):
                raise JobOptionBadValue(
                    job_id=self.ID,
                    bad_option_name=option,
                    bad_option_value=option_in,
                    condition="startswith",
                    accepted_values=option_def.get("possible_values"),
                )
            else:
                pass

        return options



#! python3  # noqa: E265

"""
    Sub-command in charge of running the main logic.

    Author: Julien M. (https://github.com/guts)
"""

# #############################################################################
# ########## Libraries #############
# ##################################

# Standard library
import argparse
import logging
from os import environ, getenv
from pathlib import Path
from sys import platform as opersys
from urllib.parse import urlsplit

# submodules
from qgis_deployment_toolbelt.constants import (
    SUPPORTED_OPERATING_SYSTEMS_CODENAMES,
    get_qdt_working_directory,
)
from qgis_deployment_toolbelt.jobs import JobsOrchestrator
from qgis_deployment_toolbelt.scenarios import ScenarioReader
from qgis_deployment_toolbelt.utils.bouncer import exit_cli_error, exit_cli_success
from qgis_deployment_toolbelt.utils.check_path import check_path
from qgis_deployment_toolbelt.utils.file_downloader import download_remote_file_to_local
from qgis_deployment_toolbelt.utils.slugger import sluggy
from qgis_deployment_toolbelt.utils.str2bool import str2bool

# #############################################################################
# ########## Globals ###############
# ##################################

logger = logging.getLogger(__name__)


# ############################################################################
# ########## FUNCTIONS ###########
# ################################


def get_remote_scenario_from_url(remote_url: str) -> Path:
    """Download remote scenario and return local file path.

    Args:
        remote_url (str): URL to remote scenario

    Returns:
        Path: local path to downloaded scenario.
    """
    # try to build file path from URL
    try:
        url_splitted = urlsplit(remote_url)
        local_filepath_for_remote_scenario = (
            "remote_scenarios/"
            f"{sluggy(url_splitted.netloc)}/"
            f"{sluggy(str(Path(url_splitted.path).parent))}/"
            f"{Path(url_splitted.path).name}"
        )
    except Exception as err:
        local_filepath_for_remote_scenario = "remote_scenarios/default/scenario.qdt.yml"
        logger.warning(
            f"Failed to extract a proper filename from URL: {remote_url}."
            f" Trace: {err}. Fallback to default: {local_filepath_for_remote_scenario}"
        )

    return download_remote_file_to_local(
        remote_url_to_download=remote_url,
        local_file_path=Path(
            get_qdt_working_directory().parent,
            local_filepath_for_remote_scenario,
        ),
        use_stream=str2bool(getenv("QDT_STREAMED_DOWNLOADS", True)),
    )


# ############################################################################
# ########## CLI #################
# ################################


def parser_main_deployment(
    subparser: argparse.ArgumentParser,
) -> argparse.ArgumentParser:
    """Set the argument parser for deployment subcommand.

    Args:
        subparser (argparse.ArgumentParser): parser to set up

    Returns:
        argparse.ArgumentParser: parser ready to use
    """
    subparser.add_argument(
        "-s",
        "--scenario",
        help="Scenario path, local or remote (HTTP/S).",
        default=getenv("QDT_SCENARIO_PATH", Path("./scenario.qdt.yml")),
        type=str,
        dest="scenario_filepath",
    )

    subparser.set_defaults(func=run)

    return subparser


# ############################################################################
# ########## MAIN ################
# ################################


def run(args: argparse.Namespace):
    """Run the main logic.

    Args:
        args (argparse.Namespace): arguments passed to the subcommand
    """
    logger.debug(f"Running {args.command} with {args}")

    # check if scenario file is local or remote
    if isinstance(args.scenario_filepath, str) and args.scenario_filepath.startswith(
        ("http",)
    ):
        args.scenario_filepath = get_remote_scenario_from_url(
            remote_url=args.scenario_filepath
        )

    # checks
    check_path(
        input_path=args.scenario_filepath,
        must_be_a_file=True,
        must_be_readable=True,
        must_exists=True,
    )

    # -- Load and validate scenario --
    try:
        scenario: ScenarioReader = ScenarioReader(in_yaml=Path(args.scenario_filepath))
    except Exception as err:
        exit_cli_error(
            f"Unable to read the scenario: {args.scenario_filepath}. Trace: {err}",
            abort=True,
        )

    # Check the validity of the scenario
    scenario_validity = scenario.validate_scenario()
    if not scenario_validity[0]:
        result_scenario_validity = (
            "Scenario validation failed. Please check the scenario file."
            "\nValidation report:\n- {}".format("\n- ".join(scenario_validity[1]))
        )
        exit_cli_error(result_scenario_validity)

    # check operating system
    if opersys not in SUPPORTED_OPERATING_SYSTEMS_CODENAMES:
        raise OSError(
            f"Your operating system {opersys} is not supported. "
            f"Supported platforms: {','.join(SUPPORTED_OPERATING_SYSTEMS_CODENAMES)}."
        )

    # -- Run --

    # Use metadata to inform which scenario is running
    logger.info(
        "Running scenario: {title} ({id}). {description}".format(**scenario.metadata)
    )

    # Set environment vars for the scenario
    for var, value in scenario.settings.items():
        if value is not None:
            logger.debug(f"Setting environment variable QDT_{var} = {value}.")
            environ[f"QDT_{var}"] = str(value)
        else:
            logger.debug(f"Ignored None value: {var}.")

    qdt_local_working_folder = get_qdt_working_directory(
        specific_value=scenario.settings.get("LOCAL_QDT_WORKDIR"),
    )
    logger.info(f"QDT working folder: {qdt_local_working_folder}")
    environ["QDT_LOCAL_QDT_WORKDIR"] = f"{qdt_local_working_folder.resolve()}"
    environ["QDT_TMP_RUNNING_SCENARIO_ID"] = str(scenario.metadata.get("id"))

    # -- STEPS JOBS
    steps_ok: list = []
    orchestrator: JobsOrchestrator = JobsOrchestrator()

    # filter out unrecognized jobs
    logger.debug("Filtering valid steps in scenario...")
    for step in scenario.steps:
        if step.get("uses") not in orchestrator.jobs_ids:
            logger.warning(f"{step.get('uses')} not found in available jobs. Skipping.")
            continue
        else:
            steps_ok.append(step)

    # run job
    for step in steps_ok:
        logger.info(f"Running step: {step.get('uses')}")
        try:
            job = orchestrator.init_job_class_from_id(
                job_id=step.get("uses"), options=step.get("with")
            )
            job.run()
        except Exception as err:
            exit_cli_error(err)

    # exit nicely
    exit_cli_success("Deployment achieved!")



#! python3  # noqa: E265

"""Helper to configure logging depending on CLI options."""

# ############################################################################
# ########## IMPORTS #############
# ################################

# standard library
import logging
from getpass import getuser
from logging.handlers import RotatingFileHandler
from os import environ, getenv
from pathlib import Path
from platform import architecture, platform, uname
from socket import gethostname

# 3rd party
import certifi
from requests.utils import DEFAULT_CA_BUNDLE_PATH

# Imports depending on operating system
if "linux" in uname().system.lower():
    import distro
else:
    distro = None

# package
from qgis_deployment_toolbelt.__about__ import __title__, __version__
from qgis_deployment_toolbelt.constants import get_qdt_logs_folder
from qgis_deployment_toolbelt.utils.proxies import get_proxy_settings
from qgis_deployment_toolbelt.utils.str2bool import str2bool

# ############################################################################
# ########## GLOBALS #############
# ################################

# logs
logger = logging.getLogger(__name__)

# ############################################################################
# ########## FUNCTIONS ###########
# ################################


def configure_logger(verbosity: int = 1, logfile: None | Path | str = None):
    """Configure logging according to verbosity from CLI.

    Args:
        verbosity (int): verbosity level
        logfile (Path, optional): file where to store log. Defaults to None.
    """
    # handle log level overridden by environment variable
    verbosity = getenv("QDT_LOGS_LEVEL", verbosity)
    try:
        verbosity = int(verbosity)
    except ValueError as err:
        logger.error(f"Bad verbosity value type: {err}. Fallback to 1.")
        verbosity = 1

    # set log level depending on verbosity argument
    if 0 < verbosity < 4:
        verbosity = 40 - (10 * verbosity)
    elif verbosity >= 4:
        # debug is the limit
        verbosity = 40 - (10 * 3)
    else:
        verbosity = 0

    # set console handler
    log_console_handler = logging.StreamHandler()
    log_console_handler.setLevel(verbosity)

    # set log file
    if not logfile:
        logging.basicConfig(
            level=verbosity,
            format="%(asctime)s||%(levelname)s||%(module)s||%(funcName)s||%(lineno)d||%(message)s",
            datefmt="%Y-%m-%d %H:%M:%S",
            handlers=[log_console_handler],
        )

    else:
        logs_folder = get_qdt_logs_folder()

        # make sure folder exists
        logs_folder.mkdir(exist_ok=True, parents=True)
        logs_filepath = Path(logs_folder, logfile)

        log_file_handler = RotatingFileHandler(
            backupCount=10,
            delay=True,
            encoding="UTF-8",
            filename=logs_filepath,
            maxBytes=3000000,
            mode="a",
        )

        # force new file by execution
        if logs_filepath.is_file():
            log_file_handler.doRollover()

        logging.basicConfig(
            level=verbosity,
            format="%(asctime)s||%(levelname)s||%(module)s||%(funcName)s||%(lineno)d||%(message)s",
            datefmt="%Y-%m-%d %H:%M:%S",
            handlers=[log_console_handler, log_file_handler],
        )

        logger.info(f"Log file: {logs_filepath}")

    headers()


def headers():
    """Basic information to log before other message."""
    # initialize the log
    logger.info(f"{'='*10} {__title__} - {__version__} {'='*10}")
    logger.debug(f"Operating System: {platform()}")
    if distro:
        logger.debug(
            f"Distribution name and version: {distro.name()} {distro.version()}"
        )
    logger.debug(f"Architecture: {architecture()[0]}")
    logger.debug(f"Computer: {gethostname()}")
    logger.debug(f"Launched by user: {getuser()}")

    if getenv("userdomain"):
        logger.debug(f"OS Domain: {getenv('userdomain')}")
    else:
        logger.debug("No OS domain detected.")

    if proxies_settings := get_proxy_settings():
        logger.debug(f"Network proxies detected: {proxies_settings}")
    else:
        logger.debug("No network proxies detected")

    # SSL CA certificates
    logger.debug(f"Installed certificate authority (CA) bundle: {certifi.where()}")
    logger.debug(f"Default certificate authority (CA) bundle: {DEFAULT_CA_BUNDLE_PATH}")
    logger.debug(
        f"Certificate authority (CA) bundle to use: {getenv('REQUESTS_CA_BUNDLE', getenv('CURL_CA_BUNDLE'))}"
    )

    if str2bool(getenv("QDT_SSL_USE_SYSTEM_STORES", False)):
        logger.debug("Option to use native system certificates stores is enabled.")
        if "REQUESTS_CA_BUNDLE" in environ:
            environ.pop("REQUESTS_CA_BUNDLE")
            logger.debug(
                "Custom path to CA Bundle (REQUESTS_CA_BUNDLE) has been removed from "
                "environment variables."
            )
        if "CURL_CA_BUNDLE" in environ:
            environ.pop("CURL_CA_BUNDLE")
            logger.debug(
                "Custom path to CA Bundle (CURL_CA_BUNDLE) has been removed from "
                "environment variables."
            )


def get_logger_filepath() -> Path | None:
    """Retrieve log filepath within logger handlers.

    Returns:
        Path | None: path to the logfile or None if no handler has baseFilename attr.
    """
    if logger.root.hasHandlers():
        for handler in logger.root.handlers:
            if hasattr(handler, "baseFilename"):
                return Path(handler.baseFilename)

    logger.warning("No file found in any log handlers.")
    return None



#! python3  # noqa E265

"""Usage from the repo root folder:

    .. code-block:: python

        # for whole test
        python -m unittest tests.test_git_handler_remote
        # for specific
        python -m unittest tests.test_git_handler_remote.TestGitHandlerRemote.test_git_url_parsed
"""

# #############################################################################
# ########## Libraries #############
# ##################################

# Standard library
import tempfile
import unittest
from pathlib import Path

# 3rd party
from dulwich.errors import NotGitRepository
from git import Repo as GitPythonRepo
from giturlparse import GitUrlParsed

# package
from qgis_deployment_toolbelt.profiles.remote_git_handler import RemoteGitHandler

# #############################################################################
# ########## Classes ###############
# ##################################


class TestGitHandlerRemote(unittest.TestCase):
    """Test module."""

    # -- Standard methods --------------------------------------------------------
    @classmethod
    def setUpClass(cls):
        """Executed when module is loaded before any test."""
        cls.good_git_url = "https://gitlab.com/Oslandia/qgis/profils_qgis_fr.git"

    # -- TESTS ---------------------------------------------------------
    def test_initialization(self):
        """Test remote git repo identifier"""
        # OK
        self.good_git_url = "https://gitlab.com/Oslandia/qgis/profils_qgis_fr.git"
        remote_git_handler = RemoteGitHandler(source_repository_url=self.good_git_url)

        self.assertEqual(remote_git_handler.SOURCE_REPOSITORY_TYPE, "git_remote")
        self.assertTrue(remote_git_handler.is_valid_git_repository())

        # KO
        bad_git_url = "https://oslandia.com"
        with self.assertRaises(NotGitRepository):
            RemoteGitHandler(bad_git_url)

    def test_is_local_git_repo(self):
        """Test local git repo identifier"""
        good_git_url = "https://github.com/octocat/Hello-World"
        git_handler = RemoteGitHandler(source_repository_url=self.good_git_url)

        with tempfile.TemporaryDirectory(
            prefix="QDT_test_check_path",
            ignore_cleanup_errors=True,
        ) as tmpdirname:
            GitPythonRepo.clone_from(url=good_git_url, to_path=Path(tmpdirname))
            # OK
            self.assertTrue(
                git_handler._is_local_path_git_repository(
                    local_path=Path(tmpdirname), raise_error=False
                )
            )
        # KO
        self.assertFalse(git_handler._is_local_path_git_repository(Path("./tests")))

    def test_git_url_parsed(self):
        """Test git parsed URL"""
        self.good_git_url = "https://gitlab.com/Oslandia/qgis/profils_qgis_fr.git"
        git_handler = RemoteGitHandler(source_repository_url=self.good_git_url)
        git_url_parsed = git_handler.url_parsed(self.good_git_url)

        # type
        self.assertIsInstance(git_url_parsed, GitUrlParsed)

        # keys
        self.assertIn("branch", git_url_parsed.data)
        self.assertIn("domain", git_url_parsed.data)
        self.assertIn("groups_path", git_url_parsed.data)
        self.assertIn("owner", git_url_parsed.data)
        self.assertIn("path", git_url_parsed.data)
        self.assertIn("path_raw", git_url_parsed.data)
        self.assertIn("pathname", git_url_parsed.data)
        self.assertIn("platform", git_url_parsed.data)
        self.assertIn("port", git_url_parsed.data)
        self.assertIn("protocol", git_url_parsed.data)
        self.assertIn("protocols", git_url_parsed.data)
        self.assertIn("repo", git_url_parsed.data)
        self.assertIn("url", git_url_parsed.data)

        # values
        self.assertIn("gitlab.com", git_url_parsed.domain)
        self.assertIn("gitlab.com", git_url_parsed.host)
        self.assertEqual("qgis", git_url_parsed.groups_path)
        self.assertEqual("Oslandia", git_url_parsed.owner)
        self.assertEqual("gitlab", git_url_parsed.platform)
        self.assertEqual("profils_qgis_fr", git_url_parsed.repo)

    def test_git_clone_remote_url(self):
        """Test git parsed URL."""
        self.good_git_url = "https://gitlab.com/Oslandia/qgis/profils_qgis_fr.git"
        git_handler = RemoteGitHandler(source_repository_url=self.good_git_url)

        with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as tmpdirname:
            local_dest = Path(tmpdirname) / "test_git_clone"
            # clone
            git_handler.download(destination_local_path=local_dest)
            # check if clone worked and new folder is a local git repo
            self.assertTrue(git_handler._is_local_path_git_repository(local_dest))

            # check pull is working
            git_handler.clone_or_pull(to_local_destination_path=local_dest)



#! python3  # noqa: E265

"""
    Utilities specific for Linux.

    Author: Julien Moura (https://github.com/guts)

"""

# #############################################################################
# ########## Libraries #############
# ##################################

# standard
import logging
import re
from functools import lru_cache
from pathlib import Path
from typing import Literal

# project
from qgis_deployment_toolbelt.utils.check_path import check_path

# #############################################################################
# ########## Globals ###############
# ##################################

logger = logging.getLogger(__name__)


# bash startup files
bash_system_dot_bash_profile: Path = Path("/etc/profile.d/qdt")
bash_user_dot_bash_profile: Path = Path.home().joinpath(".bash_profile")
bash_user_dot_bash_login: Path = Path.home().joinpath(".bash_login")
bash_user_dot_profile: Path = Path.home().joinpath(".profile")

# QDT block in-file
qdt_block_comment_start = "# BEGIN QDT MANAGED BLOCK"
qdt_block_comment_end = "# END QDT MANAGED BLOCK"


shell_path_to_names = {
    "bash": ("/bin/bash", "/usr/bin/bash"),
}

# #############################################################################
# ########## Functions #############
# ##################################


def bash_user_which_file_to_store() -> Path:
    """Determine which file to use to store environment variables.

    Returns:
        Path: path to the file to use
    """
    # check if one of the bash file exists and using it. Order matters.
    if bash_user_dot_bash_profile.is_file():
        logger.debug(
            f"Using {bash_user_dot_bash_profile} to store environment variables."
        )
        return bash_user_dot_bash_profile
    elif bash_user_dot_bash_login.is_file():
        logger.debug(
            f"Using {bash_user_dot_bash_login} to store environment variables."
        )
        return bash_user_dot_bash_login
    elif bash_user_dot_profile.is_file():
        logger.debug(f"Using {bash_user_dot_profile} to store environment variables.")
        return bash_user_dot_profile

    logger.info(
        f"None of the files exist. Creating and using {bash_user_dot_profile} to store "
        "environment variables."
    )
    bash_user_dot_profile.touch()
    return bash_user_dot_profile


@lru_cache
def find_key_from_values(value_to_find: str) -> str | None:
    """Finds the key corresponding to a given value in a dictionary where values are
    tuples.

    Args:
        value_to_find (str): The value to search for in the dictionary values.

    Returns:
        str | None: The key corresponding to the found value. Returns None if the value
            is not found.
    """

    for key, values in shell_path_to_names.items():
        if value_to_find in values:
            return key
    return None


def get_environment_variable(envvar_name: str, scope: str = "user") -> str | None:
    """Get environment variable from Linux profile file
    Args:
        envvar_name (str): environment variable name (= key) to retrieve
        scope (str, optional): environment variable scope. Must be "user" or "system",
            defaults to "user". Defaults to "user".

    Returns:
        str | None: environment variable value or None if not found
    """

    profile_file = get_profile_file(scope)

    if profile_file is not None:
        logger.debug(f"parsing {profile_file}")

        # Read file content
        with profile_file.open(mode="r", encoding="UTF-8") as file:
            lines = file.readlines()

        # look for export line
        env_value = None
        for line in lines:
            if envvar_name in line.strip():
                env_value = [item.strip() for item in re.split("=", line)][-1]

        logger.debug(f"Value for environment variable {envvar_name}: {env_value}")

        return env_value
    else:
        logger.error("Profile file was not found.")
        return None


def set_environment_variable(
    envvar_name: str, envvar_value: str | bool | int, scope: str = "user"
) -> bool:
    """Set environment variable in Linux profile file
    Args:
        envvar_name (str): environment variable name (= key) to set
        envvar_value (str): environment variable value to set
        scope (str, optional): environment variable scope. Must be "user" or "system",
            defaults to "user". Defaults to "user".

    Returns:
        bool: True if environment variable correctly set or False if not
    """

    if isinstance(envvar_value, bool):
        envvar_value = str(bool(envvar_value)).lower()

    if get_environment_variable(envvar_name, scope) is not None:
        logger.info(f"Environment variable {envvar_name} already there")
        if get_environment_variable(envvar_name, scope) != envvar_value:
            logger.info("Environment variable value to be changed")
            delete_environment_variable(envvar_name)
        else:
            logger.info(f"Environment variable already set to {envvar_value}")
            return True

    profile_file = get_profile_file(scope)

    if profile_file is not None:
        logger.debug(f"parsing {profile_file}")

        export_line = f"export {envvar_name}={str(envvar_value)}"
        block_start_found = False
        block_end_found = False
        block_end_line: int = 0
        line_found = False

        # Read file content
        with profile_file.open(mode="r", encoding="UTF-8") as file:
            lines = file.readlines()

        # look for block and export line
        line_number: int = 0
        for line in lines:
            if line.strip() == qdt_block_comment_start:
                block_start_found = True
            elif line.strip() == qdt_block_comment_end:
                block_end_found = True
                block_end_line = line_number
            elif line.strip() == export_line:
                line_found = True
            # line counter
            line_number += 1

        # check if block exist
        block_found = all([block_start_found, block_end_found])

        if line_found:
            logger.debug(
                f"Environment variable and key {envvar_name}={envvar_value} is already present"
            )
            return True
        else:
            if block_found:
                lines.insert(block_end_line, f"{export_line}\n")

                with profile_file.open(mode="w", encoding="UTF-8") as file:
                    file.writelines(lines)
                logger.info(
                    f"QDT block was already here but not the line: '{export_line}. "
                    "It has been added."
                )
                return True
            else:
                new_lines = (
                    f"\n{qdt_block_comment_start}\n",
                    f"{export_line}\n",
                    f"{qdt_block_comment_end}",
                )
                with profile_file.open(mode="a", encoding="UTF-8") as file:
                    file.writelines(new_lines)
                logger.info(
                    f"Nor QDT block and the line: '{export_line}' were present. "
                    "Both have been added."
                )
                return True
    else:
        logger.error("Profile file was not found.")
        return False


def delete_environment_variable(
    envvar_name: str, scope: Literal["system", "user"] = "user"
) -> bool:
    """Remove environment variable from Linux profile file
    Args:
        envvar_name (str): environment variable name (= key) to remove
        scope (str, optional): environment variable scope. Must be "user" or "system",
            defaults to "user". Defaults to "user".

    Returns:
        bool: True if environment variable successfully removed or False if not
    """
    # check and get profile file
    profile_file = get_profile_file(scope=scope)
    if not profile_file:
        logger.error("Profile file was not found.")
        return False

    # start parsing it
    logger.debug(f"Parsing {profile_file}")

    line_begin = f"export {envvar_name}="
    block_start_found = False
    block_start_line: int = 0
    block_end_found = False
    block_end_line: int = 0
    line_found = False
    line_begin_line: int = 0

    # read file content
    with profile_file.open(mode="r", encoding="UTF-8") as file:
        lines = file.readlines()

    # look for block and beginning of line
    line_number: int = 0
    for line in lines:
        if line.strip() == qdt_block_comment_start:
            block_start_found = True
            block_start_line = line_number
        elif line.strip() == qdt_block_comment_end:
            block_end_found = True
            block_end_line = line_number
        elif line_begin in line.strip():
            line_found = True
            line_begin_line = line_number
        # line counter
        line_number += 1

    # check if block exist
    all([block_start_found, block_end_found])

    pos = []
    if line_found:
        pos.append(line_begin_line)
        logger.debug(
            f"Environment variable and key {envvar_name} was found. "
            "It will be removed"
        )
    if block_start_found:
        pos.append(block_start_line)
        logger.debug(
            f"QDT block start was found for '{envvar_name}'. "
            "Block start will be removed."
        )
    if block_end_found:
        pos.append(block_end_line)
        logger.info(
            f"QDT block end was found for: '{envvar_name}'. "
            "Block end will be removed."
        )
    # memo: pos = [block_start_line, line_begin_line, block_end_line]
    new_lines = [lines[i] for i, e in enumerate(lines) if i not in pos]
    with profile_file.open(mode="w", encoding="UTF-8") as file:
        file.writelines(new_lines)
    logger.info(
        f"QDT block for environment variable '{envvar_name}' has been removed. "
        "Shell profile updated."
    )
    return True


def refresh_environment() -> bool:
    """Just here to be equivalent with Windows logic.

    TODO: refacto merging win32 methods in a single class.

    Returns:
        bool: always True
    """
    logger.info(
        "Changes in environment variables wont' be applied until you log out/log in or "
        f"reload your profile: 'bash source {bash_user_which_file_to_store()}'"
    )
    return True


def get_profile_file(scope: Literal["system", "user"]) -> Path | None:
    """Get Linux profile file depending on shell and scope
    Args:
        scope (str, optional): environment variable scope. Must be "user" or "system",
            defaults to "user". Defaults to "user".

    Returns:
        str | None: profile file path or None if not found
    """
    if scope == "system":
        try:
            if not bash_system_dot_bash_profile.is_file():
                bash_system_dot_bash_profile.touch()
            check_path(
                input_path=bash_system_dot_bash_profile,
                must_be_a_file=True,
                must_exists=True,
                must_be_writable=True,
                must_be_readable=True,
            )
            logger.debug(
                f"{bash_system_dot_bash_profile} exists and it's writable. It will be used."
            )
            return bash_system_dot_bash_profile
        except PermissionError as err:
            logger.error(
                f"{bash_system_dot_bash_profile} is not accessible to the user. "
                f"Trace: {err}"
            )
            return

    elif scope == "user":
        return bash_user_which_file_to_store()
    else:
        logger.error(f"Scope {scope} not handled. Fallback to 'user'.")
        return get_profile_file(scope="user")


# #############################################################################
# ##### Stand alone program ########
# ##################################

if __name__ == "__main__":
    """Standalone execution."""

    print("User profile file :", get_profile_file("user"))
    print("System profile file :", get_profile_file("system"))

    set_environment_variable("TEST_PERSISTENT_ENVIRONMENT_VARIABLE", False)
    # update_environment_variable("TEST_PERSISTENT_ENVIRONMENT_VARIABLE", False)
    # delete_environment_variable("TEST_PERSISTENT_ENVIRONMENT_VARIABLE")



#! python3  # noqa: E265

"""
    Synchronize profiles between downloaded (in QDT working folder) and installed
        profiles (in QGIS profiles'folder).

    Author: Julien Moura (https://github.com/guts)
"""


# #############################################################################
# ########## Libraries #############
# ##################################

# Standard library
import logging
from collections.abc import Iterable
from pathlib import Path
from shutil import copy2, copytree

# package
from qgis_deployment_toolbelt.jobs.generic_job import GenericJob
from qgis_deployment_toolbelt.profiles.qdt_profile import QdtProfile

# #############################################################################
# ########## Globals ###############
# ##################################

# logs
logger = logging.getLogger(__name__)


# #############################################################################
# ########## Classes ###############
# ##################################


class JobProfilesSynchronizer(GenericJob):
    """
    Job to synchronize profiles between downloaded (in QDT working folder) and installed
        profiles (in QGIS profiles'folder).
    """

    ID: str = "qprofiles-synchronizer"
    OPTIONS_SCHEMA: dict = {
        "sync_mode": {
            "type": str,
            "required": False,
            "default": "only_missing",
            "possible_values": (
                "only_missing",
                "only_different_version",
                "only_new_version",
                "overwrite",
            ),
            "condition": "in",
        },
    }
    PROFILES_NAMES_DOWNLOADED: list = []
    PROFILES_NAMES_INSTALLED: list = []

    def __init__(self, options: dict) -> None:
        """Instantiate the class.

        Args:
            options (List[dict]): list of dictionary with environment variables to set
            or remove.
        """
        super().__init__()
        self.options: dict = self.validate_options(options)

        # where QDT downloads remote repositories
        self.qdt_downloaded_repositories.mkdir(exist_ok=True, parents=True)
        logger.debug(f"Local repositories folder: {self.qdt_downloaded_repositories}")

    def run(self) -> None:
        """Execute job logic."""
        # check of there are some profiles folders within the downloaded folder
        li_qdt_profiles_from_folder = self.list_downloaded_profiles()
        if li_qdt_profiles_from_folder is None:
            logger.error("No QGIS profile found in the downloaded folder.")
            return

        # store downloaded profiles names
        self.PROFILES_NAMES_DOWNLOADED = [d.name for d in li_qdt_profiles_from_folder]
        logger.info(
            f"{len(self.PROFILES_NAMES_DOWNLOADED)} downloaded profiles: "
            f"{', '.join(self.PROFILES_NAMES_DOWNLOADED)}"
        )

        # copy profiles to the QGIS 3
        self.sync_installed_profiles_from_downloaded_profiles(
            downloaded_profiles=li_qdt_profiles_from_folder
        )

        logger.debug(f"Job {self.ID} ran successfully.")

    def compare_downloaded_with_installed_profiles(
        self, li_downloaded_profiles: Iterable[QdtProfile]
    ) -> tuple[list[QdtProfile], list[QdtProfile], list[QdtProfile]]:
        """Compare versions between downloaded (in QDT working folder) and installed
            (in QGIS) profiles.

        Args:
            li_downloaded_profiles (Iterable[QdtProfile]): downloaded profiles.

        Returns:
            tuple[list[QdtProfile], list[QdtProfile], list[QdtProfile]]: a tuple with
            the following structure (
                list of installed profiles which are outdated (lesser version number),
                list of installed profiles with a different version (lesser or greater),
                list of profiles with the same version number in downloaded/installed
                )
        """
        li_profiles_outdated = []
        li_profiles_different = []
        li_profiles_equal = []

        for downloaded_profile in li_downloaded_profiles:
            if (
                not downloaded_profile.version
                or not downloaded_profile.is_loaded_from_json
            ):
                logger.error(
                    "Unable to load profile.json from downloaded profile "
                    f"{downloaded_profile}, so it's impossible to compare "
                    "versions."
                )
                continue

            if Path(downloaded_profile.path_in_qgis, "profile.json").is_file():
                profile_installed: QdtProfile = QdtProfile.from_json(
                    profile_json_path=Path(
                        downloaded_profile.path_in_qgis, "profile.json"
                    ),
                    profile_folder=downloaded_profile.path_in_qgis,
                )
            else:
                logger.error(
                    "Unable to load profile.json from installed profile "
                    f"{downloaded_profile.path_in_qgis}, so it's impossible to compare "
                    "versions."
                )
                profile_installed: QdtProfile = QdtProfile(
                    folder=downloaded_profile.path_in_qgis, version="0.0.0"
                )

            # compare
            if profile_installed.is_older_than(downloaded_profile):
                li_profiles_outdated.append(downloaded_profile)
            elif profile_installed.version != downloaded_profile.version:
                li_profiles_different.append(downloaded_profile)
            elif profile_installed.version == downloaded_profile.version:
                li_profiles_equal.append(downloaded_profile)
            else:
                continue

        return li_profiles_outdated, li_profiles_different, li_profiles_equal

    def sync_installed_profiles_from_downloaded_profiles(
        self, downloaded_profiles: Iterable[QdtProfile]
    ) -> None:
        """Copy downloaded profiles to QGIS profiles folder. If the QGIS profiles folder
            doesn't exist, it will be created and every downloaded profile will be
            copied. If a profile is already installed, it won't be overwritten.

        Args:
            downloaded_profiles (tuple[QdtProfile]): list of downloaded profiles objects
        """
        logger.debug(f"Sync mode: {self.options.get('sync_mode')}.")
        # if local profiles folder exists or it's empty -> copy downloaded profiles
        if not self.qgis_profiles_path.exists() or not any(
            self.qgis_profiles_path.iterdir()
        ):
            logger.info(
                "The QGIS profiles folder does not exist or is empty: "
                f"{self.qgis_profiles_path.resolve()}. Probably a fresh install. "
                "Copying downloaded profiles..."
            )

            # copy all downloaded profiles
            self.sync_overwrite_local_profiles(profiles_to_copy=downloaded_profiles)

        elif self.options.get("sync_mode") == "only_missing":
            already_installed = []
            not_installed = []
            for dl_profile in downloaded_profiles:
                if not dl_profile.path_in_qgis.is_dir() or not any(
                    dl_profile.path_in_qgis.iterdir()
                ):
                    not_installed.append(dl_profile)
                else:
                    already_installed.append(dl_profile)

            if not len(not_installed):
                logger.info(
                    f"Each of the {len(downloaded_profiles)} downloaded plugins are "
                    "already installed."
                )
                compare_for_info = self.compare_downloaded_with_installed_profiles(
                    li_downloaded_profiles=downloaded_profiles
                )
                if len(compare_for_info[0]):
                    logger.info(
                        f"{compare_for_info[0]} of installed plugins are outdated comparing with"
                        f" downloaded ones: {','.join([p.name for p in compare_for_info[0]])}."
                    )
                elif len(compare_for_info[1]):
                    logger.info(
                        f"{compare_for_info[1]} of installed plugins are outdated comparing with"
                        f" downloaded ones: {','.join([p.name for p in compare_for_info[1]])}."
                    )
                else:
                    logger.info("Everything is up-to-date. Nothing to do!")
                return

            info_msg = (
                f"Mixed case. {len(not_installed)} are not installed: "
                f"{','.join([p.name for p in not_installed])}."
            )
            if len(already_installed):
                info_msg += (
                    f" {len(already_installed)} are already installed: "
                    f"{','.join([p.name for p in already_installed])}."
                )

            logger.info(info_msg)

            self.sync_overwrite_local_profiles(profiles_to_copy=not_installed)

        elif self.options.get("sync_mode") == "only_new_version":
            outdated = self.compare_downloaded_with_installed_profiles(
                li_downloaded_profiles=downloaded_profiles
            )[0]
            if not outdated:
                logger.info(
                    "All installed profiles are up-to-date with downloaded ones."
                )
                return
            self.sync_overwrite_local_profiles(profiles_to_copy=outdated)
        elif self.options.get("sync_mode") == "only_different_version":
            outdated, different, same = self.compare_downloaded_with_installed_profiles(
                li_downloaded_profiles=downloaded_profiles
            )
            if not different and not outdated:
                logger.info(
                    f"All installed profiles are the same as downloaded ones: {len(same)}"
                )
                return
            self.sync_overwrite_local_profiles(profiles_to_copy=different + outdated)
        elif self.options.get("sync_mode") == "overwrite":
            logger.debug(
                "Installed profiles are going to be overridden by downloaded ones."
            )
            # copy all downloaded profiles
            self.sync_overwrite_local_profiles(profiles_to_copy=downloaded_profiles)

        else:
            logger.debug(
                "QGIS Profiles folder already exists, it's not empty: "
                f"{self.qgis_profiles_path.resolve()}"
            )

    def sync_overwrite_local_profiles(
        self, profiles_to_copy: tuple[QdtProfile]
    ) -> None:
        """Overwrite local profiles with downloaded ones.

        Args:
            profiles_to_copy (tuple[QdtProfile]): tuple of profiles to copy
        """

        # copy downloaded profiles into this
        for d in profiles_to_copy:
            if d.path_in_qgis.exists():
                logger.info(f"Merging {d.folder} to {d.path_in_qgis}")
                installed_profile = QdtProfile(folder=d.path_in_qgis)
                d.merge_to(installed_profile)
            else:
                logger.info(f"Copying {d.folder} to {d.path_in_qgis}")
                d.path_in_qgis.mkdir(parents=True, exist_ok=True)
                copytree(
                    d.folder,
                    d.path_in_qgis,
                    copy_function=copy2,
                    dirs_exist_ok=True,
                )



#! python3  # noqa E265

"""
    Usage from the repo root folder:

    .. code-block:: bash
        # for whole tests
        python -m unittest tests.test_about
        # for specific test
        python -m unittest tests.test_about.TestAbout.test_version_semver
"""

# standard library
import unittest

# 3rd party
from packaging.version import parse
from validators import url

# project
from qgis_deployment_toolbelt import __about__

# ############################################################################
# ########## Classes #############
# ################################


class TestAbout(unittest.TestCase):
    """Test package metadata."""

    def test_metadata_types(self):
        """Test types."""
        # general
        self.assertIsInstance(__about__.__author__, str)
        self.assertIsInstance(__about__.__copyright__, str)
        self.assertIsInstance(__about__.__email__, str)
        self.assertIsInstance(__about__.__executable_name__, str)
        self.assertIsInstance(__about__.__package_name__, str)
        self.assertIsInstance(__about__.__keywords__, list)
        self.assertIsInstance(__about__.__license__, str)
        self.assertIsInstance(__about__.__summary__, str)
        self.assertIsInstance(__about__.__title__, str)
        self.assertIsInstance(__about__.__title_clean__, str)
        self.assertIsInstance(__about__.__uri_homepage__, str)
        self.assertIsInstance(__about__.__uri_repository__, str)
        self.assertIsInstance(__about__.__uri_tracker__, str)
        self.assertIsInstance(__about__.__uri__, str)
        self.assertIsInstance(__about__.__version__, str)
        self.assertIsInstance(__about__.__version_info__, tuple)

        # misc
        self.assertLessEqual(len(__about__.__title_clean__), len(__about__.__title__))

        # urls
        self.assertTrue(url(__about__.__uri_homepage__))
        self.assertTrue(url(__about__.__uri_repository__))
        self.assertTrue(url(__about__.__uri_tracker__))
        self.assertTrue(url(__about__.__uri__))

    def test_version_semver(self):
        """Test if version comply with semantic versioning."""
        self.assertTrue(parse(__about__.__version__))


# ############################################################################
# ####### Stand-alone run ########
# ################################
if __name__ == "__main__":
    unittest.main()



#! python3  # noqa: E265

"""
    Handle local git repository.

    Author: Julien Moura (https://github.com/guts).

    Inspired from: QGIS Resource Sharing
"""


# #############################################################################
# ########## Libraries #############
# ##################################

# Standard library
import logging
from pathlib import Path

# project
from qgis_deployment_toolbelt.profiles.profiles_handler_base import (
    RemoteProfilesHandlerBase,
)
from qgis_deployment_toolbelt.utils.check_path import check_path, check_var_can_be_path

# 3rd party


# #############################################################################
# ########## Globals ###############
# ##################################

# logs
logger = logging.getLogger(__name__)


# #############################################################################
# ########## Classes ###############
# ##################################
class LocalGitHandler(RemoteProfilesHandlerBase):
    """Handle local git repository."""

    def __init__(
        self,
        source_repository_path_or_uri: str | Path,
        source_repository_type: str = "git_local",
        branch_to_use: str | None = None,
    ) -> None:
        """Constructor.

        Args:
            source_repository_path_or_uri (str | Path): path to the source repository

        Raises:
            NotGitRepository: if uri_or_path doesn't point to a valid Git repository
        """
        super().__init__(
            source_repository_type=source_repository_type, branch_to_use=branch_to_use
        )

        # clean up
        if isinstance(
            source_repository_path_or_uri, str
        ) and source_repository_path_or_uri.startswith("file://"):
            source_repository_path_or_uri = source_repository_path_or_uri[7:]
            logger.debug(
                f"URI cleaning: 'file://' protocol prefix removed. Result: {source_repository_path_or_uri}"
            )
        if isinstance(
            source_repository_path_or_uri, str
        ) and source_repository_path_or_uri.endswith(".git"):
            logger.debug(
                f"URI cleaning: '.git' suffix removed. Result: {source_repository_path_or_uri}"
            )
            source_repository_path_or_uri = source_repository_path_or_uri[:-4]

        if isinstance(source_repository_path_or_uri, str) and check_var_can_be_path(
            input_var=source_repository_path_or_uri
        ):
            source_repository_path_or_uri = Path(
                source_repository_path_or_uri
            ).resolve()

        if isinstance(source_repository_path_or_uri, Path) and check_path(
            input_path=source_repository_path_or_uri,
            must_be_a_file=False,
            must_be_a_folder=True,
            must_exists=True,
            must_be_readable=True,
        ):
            source_repository_path_or_uri = source_repository_path_or_uri.resolve()

        self.SOURCE_REPOSITORY_PATH_OR_URL = source_repository_path_or_uri

        # validation
        self.is_valid_git_repository()

        self.SOURCE_REPOSITORY_ACTIVE_BRANCH = (
            self.get_active_branch_from_local_repository()
        )

        # check if passed branch exist
        if branch_to_use is None:
            self.DESTINATION_BRANCH_TO_USE = self.SOURCE_REPOSITORY_ACTIVE_BRANCH
            logger.info(
                "No branch specified. The active branch "
                f"({self.SOURCE_REPOSITORY_ACTIVE_BRANCH}) of source repository "
                f"({self.SOURCE_REPOSITORY_PATH_OR_URL}) will be used."
            )
        else:
            if not self.is_branch_existing_in_repository(branch_name=branch_to_use):
                logger.error(
                    f"Specified branch '{branch_to_use}' has not been found in source "
                    f"repository ({self.SOURCE_REPOSITORY_PATH_OR_URL}). Fallback to "
                    f"identified active branch: {self.SOURCE_REPOSITORY_ACTIVE_BRANCH}."
                )
                self.DESTINATION_BRANCH_TO_USE = self.SOURCE_REPOSITORY_ACTIVE_BRANCH
            else:
                self.DESTINATION_BRANCH_TO_USE = branch_to_use


# #############################################################################
# ##### Stand alone program ########
# ##################################

if __name__ == "__main__":
    """Standalone execution."""
    pass



#! python3  # noqa: E265

"""
    Handle remote git repository.

    Author: Julien Moura (https://github.com/guts).

    Inspired from: QGIS Resource Sharing
"""


# #############################################################################
# ########## Libraries #############
# ##################################

# Standard library
import logging
from pathlib import Path
from shutil import rmtree
from typing import Literal

# 3rd party
from dulwich import porcelain
from dulwich.errors import GitProtocolError, NotGitRepository
from dulwich.repo import Repo
from giturlparse import GitUrlParsed
from giturlparse import parse as git_parse
from giturlparse import validate as git_validate

# project
from qgis_deployment_toolbelt.utils.check_path import check_folder_is_empty

# #############################################################################
# ########## Globals ###############
# ##################################

# logs
logger = logging.getLogger(__name__)


# #############################################################################
# ########## Classes ###############
# ##################################
class RemoteProfilesHandlerBase:
    """Common git repository handler using dulwich.

    It's designed to handle thoses cases:

    - the distant repository (source) is on a local network or drive
    - the distant repository (source) is on Internet (github.com, gitlab.com, gitlab.company.org...)
    - the local repository (destination) is on a local network or drive
    """

    SOURCE_REPOSITORY_ACTIVE_BRANCH: str | None = None
    SOURCE_REPOSITORY_PATH_OR_URL: Path | str | None = None
    SOURCE_REPOSITORY_TYPE: (
        Literal["git_local", "git_remote", "http", "local", "remote"] | None
    ) = None

    DESTINATION_PATH: Path | None = None
    DESTINATION_BRANCH_TO_USE: str | None = None

    def __init__(
        self,
        source_repository_type: Literal[
            "git_local", "git_remote", "http", "local", "remote"
        ],
        branch_to_use: str | None = None,
    ) -> None:
        """Object instanciation.

        Args:
            source_repository_type (Literal["git_local", "git_remote", "local", "remote"]): type of source
                repository
            branch_to_use (str | None, optional): branch to clone or checkout. If None,
                the source active branch will be used. Defaults to None.
        """
        self.DESTINATION_BRANCH_TO_USE = branch_to_use
        self.SOURCE_REPOSITORY_TYPE = source_repository_type

    def url_parsed(self, remote_git_url: str) -> GitUrlParsed:
        """Return URL parsed to extract git information.

        Returns:
            GitUrlParsed: parsed URL object
        """
        return git_parse(remote_git_url)

    def is_valid_git_repository(
        self,
        source_repository_path_or_url: Path | str | None = None,
        force_type: Literal["git_local", "git_remote", "local", "remote"] | None = None,
        raise_error: bool = True,
    ) -> bool:
        """Determine if the given path or URL is a valid repository or not.

        Args:
            source_repository_path_or_url (Path | str | None, optional): _description_.
                Defaults to None.
            force_type (Literal["git_local","local", "remote"], optional): force git repository
                type to check. If None, it uses the SOURCE_REPOSITORY_TYPE attribute.
                Defaults None.
            raise_error (bool, optional): if True, it raises an exception. Defaults
                to True.

        Raises:
            NotGitRepository: if given path or URL is not a valid Git repository

        Returns:
            bool: True if the given path or URL is a valid Git repository
        """
        valid_source = True

        # if no local git repository passed, try to use URL defined at object level
        if source_repository_path_or_url is None and isinstance(
            self.SOURCE_REPOSITORY_PATH_OR_URL, (Path, str)
        ):
            source_repository_path_or_url: str | Path = (
                self.SOURCE_REPOSITORY_PATH_OR_URL
            )
            logger.info(
                f"Using source repository set at object's level: {source_repository_path_or_url}"
            )

        # use the repository type if forced or attribute
        if force_type is None:
            repository_type = self.SOURCE_REPOSITORY_TYPE
        else:
            repository_type = force_type

        # check according to the repository types
        if repository_type in (
            "git_remote",
            "remote",
        ) and not self._is_url_git_repository(
            remote_git_url=source_repository_path_or_url
        ):
            valid_source = False
        elif repository_type in (
            "git_local",
            "local",
        ) and not self._is_local_path_git_repository(
            local_path=source_repository_path_or_url
        ):
            valid_source = False

        # log and/or raise
        err_message = f"{source_repository_path_or_url} is not a valid repository."
        if not valid_source and raise_error:
            raise NotGitRepository(
                f"{source_repository_path_or_url} is not a valid repository."
            )
        elif not valid_source:
            logger.debug(err_message)
        else:
            logger.debug(
                f"{source_repository_path_or_url} is a valid "
                f"{self.SOURCE_REPOSITORY_TYPE} repository."
            )

        return valid_source

    def _is_local_path_git_repository(
        self, local_path: Path | None, raise_error: bool = True
    ) -> bool:
        """Check if local folder is a git repository.

        Args:
            local_path (Path, optional): path to check
            raise_error (bool, optional): if enabled, log message is an error, debug
                if not. Defaults to True.

        Returns:
            bool: True if there is a .git subfolder
        """
        # if no local git repository passed, try to use URL defined at object level
        if local_path is None and isinstance(self.SOURCE_REPOSITORY_PATH_OR_URL, Path):
            local_path: Path = self.SOURCE_REPOSITORY_PATH_OR_URL

        try:
            Repo(root=f"{local_path.resolve()}")
            return True
        except NotGitRepository as err:
            if not raise_error:
                logger.debug(f"{local_path} is not a valid Git repository")
                return False
            logger.error(f"{local_path} is not a valid Git repository. Trace: {err}")
            return False
        except Exception as err:
            logger.error(
                f"Something went wrong when checking if {local_path} is a valid "
                f"Git repository. Trace: {err}"
            )
            return False

    def _is_url_git_repository(
        self, remote_git_url: str | None = None, raise_error: bool = True
    ) -> bool:
        """Check if the remote URL is a git repository.

        Args:
            remote_git_url (str): URL pointing to a remote git repository.
            raise_error (bool, optional): if enabled, log message is an error, debug
                if not. Defaults to True.

        Returns:
            bool: True if the URL is a valid git repository.
        """
        # if remote git URL not passed, try to use URL defined at object level
        if remote_git_url is None and isinstance(
            self.SOURCE_REPOSITORY_PATH_OR_URL, str
        ):
            remote_git_url = self.SOURCE_REPOSITORY_PATH_OR_URL

        try:
            return git_validate(remote_git_url)
        except Exception as err:
            if not raise_error:
                logger.debug(f"{remote_git_url} is not a valid Git repository")
                return False
            logger.error(
                f"{remote_git_url} is not a valid Git repository. Trace: {err}"
            )
            return False

    def get_active_branch_from_local_repository(
        self, local_git_repository_path: Path | None = None
    ) -> str:
        """Retrieve git active branch from a local repository. Mainly a checker and a
            wrapper around dulwich logic.

        Args:
            local_git_repository_path (Path | None, optional): path to the local
                repository. If not defined, the SOURCE_REPOSITORY_PATH_OR_URL object's
                attribute is used if it exists. Defaults to None.

        Raises:
            NotGitRepository: if the path is not a valid Git Repository

        Returns:
            str: branch name
        """
        # if no local git repository passed, try to use URL defined at object level
        if local_git_repository_path is None and isinstance(
            self.SOURCE_REPOSITORY_PATH_OR_URL, Path
        ):
            local_git_repository_path: Path = self.SOURCE_REPOSITORY_PATH_OR_URL

        self.is_valid_git_repository(
            source_repository_path_or_url=local_git_repository_path
        )

        return porcelain.active_branch(
            repo=Repo(root=f"{local_git_repository_path.resolve()}")
        ).decode()

    def is_branch_existing_in_repository(
        self,
        branch_name: str | bytes,
        repository_path_or_url: Path | str | None = None,
    ) -> bool:
        """Determine if the given branch name is part of the given repository.

        Args:
            branch_name (str | bytes): _description_
            repository_path_or_url (Path | str, optional): URL or path pointing to a
                git repository. If None, it uses the SOURCE_REPOSITORY_PATH_OR_URL
                object's attribute

        Raises:
            NotGitRepository: if the path is not a valid Git Repository

        Returns:
            bool: True is the rbanch is part of given repository existing branches
        """
        # make sure this is string
        if isinstance(branch_name, bytes):
            branch_name = branch_name.decode()

        # if no local git repository passed, try to use URL defined at object level
        if repository_path_or_url is None and isinstance(
            self.SOURCE_REPOSITORY_PATH_OR_URL, (Path, str)
        ):
            repository_path_or_url: Path | str = self.SOURCE_REPOSITORY_PATH_OR_URL

        # check if URL or path is pointing to a valid git repository
        if not self.is_valid_git_repository(
            source_repository_path_or_url=repository_path_or_url
        ):
            raise NotGitRepository(
                f"{repository_path_or_url} is not a valid repository."
            )

        # clean branch name
        refs_heads_prefix = "refs/heads/"
        branch_name = branch_name.removeprefix(refs_heads_prefix)

        return branch_name in [
            branch.removeprefix(refs_heads_prefix)
            for branch in self.list_remote_branches(
                source_repository_path_or_url=repository_path_or_url
            )
        ]

    def list_remote_branches(
        self, source_repository_path_or_url: Path | str | None = None
    ) -> tuple[str]:
        """Retrieve git active branch from a local repository. Mainly a checker and a
            wrapper around dulwich logic.

        Args:
            source_repository_path_or_url (Path | str, optional): URL or path pointing
                to a git repository.

        Raises:
            NotGitRepository: if the path is not a valid Git Repository

        Returns:
            tuple[str]: tuple of branch complete names \
                ('refs/heads/profile-for-qgis-3-34', 'refs/heads/main')
        """
        # if no local git repository passed, try to use URL defined at object level
        if source_repository_path_or_url is None and isinstance(
            self.SOURCE_REPOSITORY_PATH_OR_URL, (Path, str)
        ):
            source_repository_path_or_url: Path | str = (
                self.SOURCE_REPOSITORY_PATH_OR_URL
            )

        # check if URL or path is pointing to a valid git repository
        if not self.is_valid_git_repository(
            source_repository_path_or_url=source_repository_path_or_url
        ):
            raise NotGitRepository(
                f"{source_repository_path_or_url} is not a valid repository."
            )

        ls_remote_refs: dict = porcelain.ls_remote(
            remote=f"{source_repository_path_or_url}"
        )
        if isinstance(ls_remote_refs, dict):
            source_repository_branches: list[str] = [
                ref.decode() for ref in ls_remote_refs if ref.startswith(b"refs/heads/")
            ]
            logger.debug(
                f"{len(source_repository_branches)} branche(s) found in repository "
                f"{source_repository_path_or_url}: "
                f"{' ; '.join(source_repository_branches)}"
            )
            return tuple(source_repository_branches)
        else:
            return ("",)

    def download(self, destination_local_path: Path) -> Repo:
        """Generic wrapper around the specific logic of this handler.

        Args:
            destination_local_path (Path): path to the local folder where to download

        Returns:
            Repo: the local repository object
        """
        if isinstance(destination_local_path, Path):
            destination_local_path = destination_local_path.resolve()

        local_git_repository = self.clone_or_pull(
            to_local_destination_path=destination_local_path
        )

        if isinstance(local_git_repository, Repo):
            self.DESTINATION_BRANCH_TO_USE = porcelain.active_branch(
                local_git_repository
            )

        return local_git_repository

    def clone_or_pull(self, to_local_destination_path: Path, attempt: int = 1) -> Repo:
        """Clone or fetch/pull remote repository to local path. If this one doesn't exist,
        it's created. If fetch or pull action fail, it removes the existing folder and
        clone the remote again.

        Args:
            to_local_destination_path (Path): path to the folder where to clone (or pull)
            attempt (int): attempt count. If attempts < 2 and it fails, the destination
                path is completely removed before cloning again. Defaults to 1.

        Raises:
            err: if something fails during clone or pull operations

        Returns:
            Repo: the local repository object
        """
        # clone
        if (
            not to_local_destination_path.exists()
            or check_folder_is_empty(to_local_destination_path)
        ) and not self.is_valid_git_repository(
            source_repository_path_or_url=to_local_destination_path,
            raise_error=False,
            force_type="git_local",
        ):
            try:
                logger.debug("Start cloning operations...")
                return self._clone(local_path=to_local_destination_path)
            except Exception as err:
                logger.error(
                    "Error cloning the source repository "
                    f"{self.SOURCE_REPOSITORY_PATH_OR_URL} "
                    f"(branch {self.SOURCE_REPOSITORY_ACTIVE_BRANCH}) "
                    f"to {to_local_destination_path}. "
                    f"Trace: {err}."
                )
                if attempt < 2:
                    logger.error(
                        "Clone fail 1/2. Removing target folder and trying again..."
                    )
                    rmtree(path=to_local_destination_path, ignore_errors=True)
                    return self.clone_or_pull(
                        to_local_destination_path=to_local_destination_path, attempt=2
                    )
                logger.critical("Clone fail 2/2. Abort.")
                rmtree(path=to_local_destination_path, ignore_errors=True)
                raise err
        elif to_local_destination_path.exists() and self.is_valid_git_repository(
            source_repository_path_or_url=to_local_destination_path,
            raise_error=False,
            force_type="git_local",
        ):
            # FETCH
            logger.debug("Start fetching operations...")
            try:
                self._fetch(local_path=to_local_destination_path)
            except GitProtocolError as error:
                logger.error(
                    "Error fetching source repository "
                    f"{self.SOURCE_REPOSITORY_PATH_OR_URL} "
                    f"to {to_local_destination_path.resolve()}. Trace: {error}."
                )
                rmtree(path=to_local_destination_path, ignore_errors=True)
                return self.clone_or_pull(
                    to_local_destination_path=to_local_destination_path
                )
            # PULL
            logger.debug("Start pulling operations...")
            try:
                return self._pull(local_path=to_local_destination_path)
            except GitProtocolError as error:
                logger.error(
                    f"Error pulling {self.SOURCE_REPOSITORY_PATH_OR_URL} "
                    f"repository to {to_local_destination_path.resolve()}. Trace: {error}."
                    "Trying to remove the local folder and cloning again..."
                )
                rmtree(path=to_local_destination_path, ignore_errors=True)
                return self.clone_or_pull(
                    to_local_destination_path=to_local_destination_path
                )
        elif not to_local_destination_path.exists():
            logger.debug(
                f"Local path does not exists: {to_local_destination_path.as_uri()}. "
                "Creating it and trying again..."
            )
            to_local_destination_path.mkdir(parents=True, exist_ok=True)
            return self.clone_or_pull(
                to_local_destination_path=to_local_destination_path
            )
        else:
            logger.critical(
                f"Case not handle. Context: {to_local_destination_path} "
                f"(empty={check_folder_is_empty(to_local_destination_path)}) "
                f"valid repo={self.is_valid_git_repository(source_repository_path_or_url=to_local_destination_path, raise_error=False)}"
            )
            return None

    def _clone(self, local_path: Path) -> Repo:
        """Clone the remote repository to local path.

        Args:
            local_path (Path): path to the folder where to clone

        Returns:
            Repo: the local repository object
        """
        # make sure folder and its parents exist
        if not local_path.exists():
            local_path.mkdir(parents=True, exist_ok=True)

        # make sure branch is bytes
        if isinstance(self.DESTINATION_BRANCH_TO_USE, str) and len(
            self.DESTINATION_BRANCH_TO_USE
        ):
            branch = self.DESTINATION_BRANCH_TO_USE.encode()
        elif isinstance(self.DESTINATION_BRANCH_TO_USE, bytes):
            branch = self.DESTINATION_BRANCH_TO_USE
        else:
            branch = None

        logger.debug(
            f"Cloning repository {self.SOURCE_REPOSITORY_PATH_OR_URL} ({branch=}) to {local_path}"
        )

        if self.SOURCE_REPOSITORY_TYPE in ("git_local", "local"):
            with porcelain.open_repo_closing(
                path_or_repo=self.SOURCE_REPOSITORY_PATH_OR_URL
            ) as repo_obj:
                repo_obj.clone(
                    target_path=f"{local_path.resolve()}",
                    branch=branch,
                    mkdir=False,
                    checkout=True,
                    progress=None,
                )
        elif self.SOURCE_REPOSITORY_TYPE in ("git_remote", "remote"):
            repo_obj = porcelain.clone(
                source=self.SOURCE_REPOSITORY_PATH_OR_URL,
                target=f"{local_path.resolve()}",
                branch=branch,
            )
        else:
            raise NotImplementedError(f"{self.SOURCE_REPOSITORY_TYPE} is not supported")

        gobj = repo_obj.get_object(repo_obj.head())
        logger.debug(
            f"Active branch: {porcelain.active_branch(repo_obj)}. "
            f"Latest commit cloned: {gobj.sha().hexdigest()} by {gobj.author}"
            f" at {gobj.commit_time}."
        )
        return repo_obj

    def _fetch(self, local_path: Path) -> Repo:
        """Fetch the remote repository from the existing local repository.

        Args:
            local_path (Path): path to the folder where to fetch

        Returns:
            Repo: the local repository object
        """
        # if source repository is a local path, let's convert it into str
        if isinstance(self.SOURCE_REPOSITORY_PATH_OR_URL, Path):
            source_repository = f"{self.SOURCE_REPOSITORY_PATH_OR_URL.resolve()}"
        else:
            source_repository = str(self.SOURCE_REPOSITORY_PATH_OR_URL)

        # with porcelain.open_repo_closing(str(local_path.resolve())) as local_repo:
        logger.info(
            f"Fetching repository {source_repository} to {local_path}",
        )

        destination_local_repository = Repo(root=f"{local_path.resolve()}")
        porcelain.fetch(
            repo=destination_local_repository,
            remote_location=source_repository,
            force=True,
            prune=True,
            prune_tags=True,
        )
        destination_local_repository.close()

        logger.debug(
            f"Repository {local_path.resolve()} has been fetched from "
            f"remote {source_repository}. "
            f"Local active branch: {porcelain.active_branch(destination_local_repository)}."
        )

        return destination_local_repository

    def _pull(self, local_path: Path) -> Repo:
        """Pull the remote repository from the existing local repository.

        Args:
            local_path (Path): path to the folder where to pull

        Returns:
            Repo: the local repository object
        """
        # if source repository is a local path, let's convert it into str
        if isinstance(self.SOURCE_REPOSITORY_PATH_OR_URL, Path):
            source_repository = f"{self.SOURCE_REPOSITORY_PATH_OR_URL.resolve()}"
        else:
            source_repository = str(self.SOURCE_REPOSITORY_PATH_OR_URL)

        logger.info(f"Pulling repository {source_repository} to {local_path}")

        destination_local_repository = Repo(root=f"{local_path.resolve()}")
        porcelain.pull(
            repo=local_path,
            remote_location=source_repository,
            force=True,
        )
        gobj = destination_local_repository.get_object(
            destination_local_repository.head()
        )
        logger.debug(
            f"Repository {local_path.resolve()} has been pulled. "
            f"Local active branch: {porcelain.active_branch(destination_local_repository)}. "
            f"Latest commit cloned: {gobj.sha().hexdigest()} by {gobj.author}"
            f" at {gobj.commit_time}"
        )

        destination_local_repository.close()
        return destination_local_repository


# #############################################################################
# ##### Stand alone program ########
# ##################################

if __name__ == "__main__":
    """Standalone execution."""
    pass
